# -*- coding: utf-8 -*-

# Copyright 2016-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Collection of extractors for various imagehosts"""

from .common import Extractor, Message
from .. import text, exception
from ..cache import memcache


class ImagehostImageExtractor(Extractor):
    """Base class for single-image extractors for various imagehosts"""
    basecategory = "imagehost"
    subcategory = "image"
    archive_fmt = "{token}"
    parent = True
    _params = None
    _cookies = None
    _encoding = None
    _validate = None

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.page_url = (self.root or "https://") + match[1]
        self.token = match[2]

        if self._params == "simple":
            self._params = {
                "imgContinue": "Continue+to+image+...+",
            }
        elif self._params == "complex":
            self._params = {
                "op": "view",
                "id": self.token,
                "pre": "1",
                "adb": "1",
                "next": "Continue+to+image+...+",
            }

    def items(self):
        page = self.request(
            self.page_url,
            method=("POST" if self._params else "GET"),
            data=self._params,
            cookies=self._cookies,
            encoding=self._encoding,
        ).text

        url, filename = self.get_info(page)
        if not url:
            return

        if filename:
            data = text.nameext_from_name(filename)
            if not data["extension"]:
                data["extension"] = text.ext_from_url(url)
        else:
            data = text.nameext_from_url(url)
        data["token"] = self.token
        data["post_url"] = self.page_url
        data.update(self.metadata(page))

        if url.startswith("http:"):
            url = "https:" + url[5:]
        if self._validate is not None:
            data["_http_validate"] = self._validate

        yield Message.Directory, "", data
        yield Message.Url, url, data

    def get_info(self, page):
        """Find image-url and string to get filename from"""

    def metadata(self, page):
        """Return additional metadata"""
        return ()

    def not_found(self, resource=None):
        raise exception.NotFoundError(resource or self.__class__.subcategory)


class ImxtoImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from imx.to"""
    category = "imxto"
    pattern = (r"(?:https?://)?(?:www\.)?((?:imx\.to|img\.yt)"
               r"/(?:i/|img-)(\w+)(\.html)?)")
    example = "https://imx.to/i/ID"
    _params = "simple"
    _encoding = "utf-8"

    def __init__(self, match):
        ImagehostImageExtractor.__init__(self, match)
        if "/img-" in self.page_url:
            self.page_url = self.page_url.replace("img.yt", "imx.to")

    def get_info(self, page):
        url, pos = text.extract(
            page, '<div style="text-align:center;"><a href="', '"')
        if not url:
            self.not_found()
        filename, pos = text.extract(page, ' title="', '"', pos)
        return url, filename or None

    def metadata(self, page):
        extr = text.extract_from(page, page.index("[ FILESIZE <"))
        size = extr(">", "</span>").replace(" ", "")[:-1]
        width, _, height = extr(">", " px</span>").partition("x")
        return {
            "size"  : text.parse_bytes(size),
            "width" : text.parse_int(width),
            "height": text.parse_int(height),
            "hash"  : extr(">", "</span>"),
        }


class ImxtoGalleryExtractor(ImagehostImageExtractor):
    """Extractor for image galleries from imx.to"""
    category = "imxto"
    subcategory = "gallery"
    pattern = r"(?:https?://)?(?:www\.)?(imx\.to/g/([^/?#]+))"
    example = "https://imx.to/g/ID"

    def items(self):
        page = self.request(self.page_url).text
        title, pos = text.extract(page, '<div class="title', '<')
        data = {
            "_extractor": ImxtoImageExtractor,
            "title": text.unescape(title.partition(">")[2]).strip(),
        }

        params = {"page": 1}
        while True:
            for url in text.extract_iter(page, "<a href=", " ", pos):
                if "/i/" in url:
                    yield Message.Queue, url.strip("\"'"), data

            if 'class="pagination' not in page or \
                    'class="disabled">Last' in page:
                return

            params["page"] += 1
            page = self.request(self.page_url, params=params).text


class AcidimgImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from acidimg.cc"""
    category = "acidimg"
    pattern = r"(?:https?://)?((?:www\.)?acidimg\.cc/img-([a-z0-9]+)\.html)"
    example = "https://acidimg.cc/img-abc123.html"
    _params = "simple"
    _encoding = "utf-8"

    def get_info(self, page):
        url, pos = text.extract(page, "<img class='centred' src='", "'")
        if not url:
            url, pos = text.extract(page, '<img class="centred" src="', '"')
            if not url:
                self.not_found()

        filename, pos = text.extract(page, "alt='", "'", pos)
        if not filename:
            filename, pos = text.extract(page, 'alt="', '"', pos)

        return url, filename or None


class ImagevenueImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from imagevenue.com"""
    category = "imagevenue"
    pattern = (r"(?:https?://)?((?:www|img\d+)\.imagevenue\.com"
               r"/([A-Z0-9]{8,10}|view/.*|img\.php\?.*))")
    example = "https://www.imagevenue.com/ME123456789"

    @property
    @memcache(maxage=3*3600)
    def _cookies(self):
        return self.request(self.page_url).cookies

    def get_info(self, page):
        try:
            pos = page.index('class="card-body')
        except ValueError:
            self.not_found()

        url, pos = text.extract(page, '<img src="', '"', pos)
        if url.endswith("/loader.svg"):
            url, pos = text.extract(page, '<img src="', '"', pos)
        filename, pos = text.extract(page, 'alt="', '"', pos)
        return url, text.unescape(filename)

    def _validate(self, response):
        hget = response.headers.get
        return not (
            hget("content-length") == "14396" and
            hget("content-type") == "image/jpeg" and
            hget("last-modified") == "Mon, 04 May 2020 07:19:52 GMT"
        )


class ImagetwistImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from imagetwist.com"""
    category = "imagetwist"
    pattern = (r"(?:https?://)?((?:www\.|phun\.)?image(?:twist|haha)\.com"
               r"/([a-z0-9]{12}))")
    example = "https://imagetwist.com/123456abcdef/NAME.EXT"

    @property
    @memcache(maxage=3*3600)
    def _cookies(self):
        return self.request(self.page_url).cookies

    def get_info(self, page):
        url     , pos = text.extract(page, '<img src="', '"')
        if url and url.startswith("/imgs/"):
            self.not_found()
        filename, pos = text.extract(page, ' alt="', '"', pos)
        return url, filename


class ImagetwistGalleryExtractor(ImagehostImageExtractor):
    """Extractor for galleries from imagetwist.com"""
    category = "imagetwist"
    subcategory = "gallery"
    pattern = (r"(?:https?://)?((?:www\.|phun\.)?image(?:twist|haha)\.com/("
               r"p/[^/?#]+/(\d+)|"
               r"\?[^#]*\bfld_id=\d+[^#]*&page=\d+))")
    example = "https://imagetwist.com/p/USER/12345/TITLE"

    def items(self):
        url = self.page_url
        root = url[:url.find("/", 8)]
        page = self.request(url).text

        extr = text.extract_from(page)
        data = {
            "_extractor"   : ImagetwistImageExtractor,
            "gallery_title": text.unescape(extr('page_main_title">', "<")),
            "gallery_id"   : self.groups[2] or extr("&amp;fld_id=", "&"),
        }
        del extr

        while True:
            gallery = text.extr(page, 'class="gallerys', "</div")
            for path in text.extract_iter(gallery, ' href="', '"'):
                yield Message.Queue, root + path, data

            pos = page.find("&#187;</a>")
            if pos < 0:
                break
            qs = text.unescape(text.rextr(page, "href='", "'", pos))

            page = self.request(f"{root}/{qs}").text


class ImgadultImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from imgadult.com"""
    category = "imgadult"
    _cookies = {"img_i_d": "1"}
    pattern = r"(?:https?://)?((?:www\.)?imgadult\.com/img-([0-9a-f]+)\.html)"
    example = "https://imgadult.com/img-0123456789abc.html"

    def get_info(self, page):
        url , pos = text.extract(page, "' src='", "'")
        name, pos = text.extract(page, "alt='", "'", pos)

        if name:
            name, _, rhs = name.rpartition(" image hosted at ImgAdult.com")
            if not name:
                name = rhs
            name = text.unescape(name)

        return url, name


class ImgspiceImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from imgspice.com"""
    category = "imgspice"
    pattern = r"(?:https?://)?((?:www\.)?imgspice\.com/([^/?#]+))"
    example = "https://imgspice.com/ID/NAME.EXT.html"

    def get_info(self, page):
        pos = page.find('id="imgpreview"')
        if pos < 0:
            self.not_found()
        url , pos = text.extract(page, 'src="', '"', pos)
        name, pos = text.extract(page, 'alt="', '"', pos)
        return url, text.unescape(name)


class PixhostImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from pixhost.to"""
    category = "pixhost"
    root = "https://pixhost.to"
    pattern = (r"(?:https?://)?(?:www\.)?pixhost\.(?:to|org)"
               r"(/show/\d+/(\d+)_[^/?#]+)")
    example = "https://pixhost.to/show/123/12345_NAME.EXT"
    _cookies = {"pixhostads": "1", "pixhosttest": "1"}

    def get_info(self, page):
        self.kwdict["directory"] = self.page_url.rsplit("/")[-2]
        url , pos = text.extract(page, "class=\"image-img\" src=\"", "\"")
        name, pos = text.extract(page, "alt=\"", "\"", pos)
        return url, text.unescape(name) if name else None


class PixhostGalleryExtractor(ImagehostImageExtractor):
    """Extractor for image galleries from pixhost.to"""
    category = "pixhost"
    subcategory = "gallery"
    root = "https://pixhost.to"
    pattern = (r"(?:https?://)?(?:www\.)?pixhost\.(?:to|org)"
               r"(/gallery/([^/?#]+))")
    example = "https://pixhost.to/gallery/ID"

    def items(self):
        page = text.extr(self.request(
            self.page_url).text, 'class="images"', "</div>")
        data = {"_extractor": PixhostImageExtractor}
        for url in text.extract_iter(page, '<a href="', '"'):
            yield Message.Queue, url, data


class PostimgImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from postimages.org"""
    category = "postimg"
    root = "https://postimg.cc"
    pattern = (r"(?:https?://)?(?:www\.)?(?:postim(?:ages|g)|pixxxels)"
               r"\.(?:cc|org)(/(?!gallery/)(?:image/)?([^/?#]+)/?)")
    example = "https://postimg.cc/ID"

    def get_info(self, page):
        pos = page.index(' id="download"')
        url     , pos = text.rextract(page, ' href="', '"', pos)
        filename, pos = text.extract(page, ' alt="', '"', pos)
        return url, text.unescape(filename) if filename else None


class PostimgGalleryExtractor(ImagehostImageExtractor):
    """Extractor for images galleries from postimages.org"""
    category = "postimg"
    subcategory = "gallery"
    root = "https://postimg.cc"
    pattern = (r"(?:https?://)?(?:www\.)?(?:postim(?:ages|g)|pixxxels)"
               r"\.(?:cc|org)(/gallery/([^/?#]+))")
    example = "https://postimg.cc/gallery/ID"

    def items(self):
        page = self.request(self.page_url).text
        title = text.unescape(text.extr(
            page, 'property="og:title" content="', ' — Postimages"'))

        url = self.root + "/json"
        params = {
            "action": "list",
            "page"  : 1,
            "album" : self.groups[1],
        }

        base = self.root + "/"
        while True:
            data = self.request_json(url, params=params)

            for token, t, name, ext, w, h, _, _, _, _, _ in data["images"]:
                yield Message.Queue, base + t, {
                    "_extractor"   : PostimgImageExtractor,
                    "gallery_title": title,
                    "token"    : token,
                    "filename" : name,
                    "extension": ext,
                    "width"    : w,
                    "height"   : h,
                    "thumbnail": t,
                }

            if not data.get("has_page_next"):
                break
            params["page"] += 1


class TurboimagehostImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from www.turboimagehost.com"""
    category = "turboimagehost"
    pattern = (r"(?:https?://)?((?:www\.)?turboimagehost\.com"
               r"/p/(\d+)/[^/?#]+\.html)")
    example = "https://www.turboimagehost.com/p/12345/NAME.EXT.html"

    def get_info(self, page):
        url = text.extract(page, 'src="', '"', page.index("<img "))[0]
        return url, None


class TurboimagehostGalleryExtractor(ImagehostImageExtractor):
    """Extractor for image galleries from turboimagehost.com"""
    category = "turboimagehost"
    subcategory = "gallery"
    pattern = (r"(?:https?://)?((?:www\.)?turboimagehost\.com"
               r"/album/(\d+)/([^/?#]*))")
    example = "https://www.turboimagehost.com/album/12345/GALLERY_NAME"

    def items(self):
        data = {"_extractor": TurboimagehostImageExtractor}
        params = {"p": 1}

        while True:
            page = self.request(self.page_url, params=params).text

            if params["p"] == 1 and \
                    "Requested gallery don`t exist on our website." in page:
                self.not_found()

            thumb_url = None
            for thumb_url in text.extract_iter(page, '"><a href="', '"'):
                yield Message.Queue, thumb_url, data
            if thumb_url is None:
                return

            params["p"] += 1


class ViprImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from vipr.im"""
    category = "vipr"
    pattern = r"(?:https?://)?(vipr\.im/(\w+))"
    example = "https://vipr.im/abc123.html"

    def get_info(self, page):
        url, pos = text.extract(page, '<img src="', '"')
        if not url or url[0] != "h":
            self.not_found()
        alt, pos = text.extract(page, ' alt="', '"', pos)
        return url, alt and text.unescape(alt)


class ImgclickImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from imgclick.net"""
    category = "imgclick"
    pattern = r"(?:https?://)?((?:www\.)?imgclick\.net/([^/?#]+))"
    example = "http://imgclick.net/abc123/NAME.EXT.html"
    _params = "complex"

    def get_info(self, page):
        url     , pos = text.extract(page, '<br><img src="', '"')
        filename, pos = text.extract(page, 'alt="', '"', pos)
        return url, filename


class FappicImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from fappic.com"""
    category = "fappic"
    pattern = (r"(?:https?://)?(?:www\.|img\d+\.)?fappic\.com"
               r"/(?:i/\d+/())?(\w{10,})(?:/|\.)\w+")
    example = "https://fappic.com/abcde12345/NAME.EXT"

    def __init__(self, match):
        Extractor.__init__(self, match)

        thumb, token = self.groups
        if thumb is not None and token.endswith("_t"):
            self.token = token = token[:-2]
        else:
            self.token = token
        self.page_url = f"https://fappic.com/{token}/pic.jpg"

    def get_info(self, page):
        url     , pos = text.extract(page, '<a href="#"><img src="', '"')
        filename, pos = text.extract(page, 'alt="', '"', pos)
        return url, text.re(r"^Porn[ -]Pic(?:s|ture)[ -]").sub("", filename)


class PicstateImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from picstate.com"""
    category = "picstate"
    pattern = r"(?:https?://)?((?:www\.)?picstate\.com/view/full/([^/?#]+))"
    example = "https://picstate.com/view/full/123"

    def get_info(self, page):
        pos = page.index(' id="image_container"')
        url     , pos = text.extract(page, '<img src="', '"', pos)
        filename, pos = text.extract(page, 'alt="', '"', pos)
        return url, filename


class ImgdriveImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from imgdrive.net"""
    category = "imgdrive"
    pattern = (r"(?:https?://)?(?:www\.)?(img(drive|taxi|wallet)\.(?:com|net)"
               r"/img-(\w+)\.html)")
    example = "https://imgdrive.net/img-0123456789abc.html"

    def __init__(self, match):
        path, category, self.token = match.groups()
        self.page_url = "https://" + path
        self.category = "img" + category
        Extractor.__init__(self, match)

    def get_info(self, page):
        title, pos = text.extract(
            page, 'property="og:title" content="', '"')
        image, pos = text.extract(
            page, 'property="og:image" content="', '"', pos)
        return image.replace("/small/", "/big/"), title.rsplit(" | ", 2)[0]


class SilverpicImageExtractor(ImagehostImageExtractor):
    """Extractor for single images from silverpic.com"""
    category = "silverpic"
    root = "https://silverpic.net"
    _params = "complex"
    pattern = (r"(?:https?://)?(?:www\.)?silverpic\.(?:net|com)"
               r"(/([a-z0-9]{10,})/[\S]+\.html)")
    example = "https://silverpic.net/a1b2c3d4f5g6/NAME.EXT.html"

    def get_info(self, page):
        url, pos = text.extract(page, '<img src="/img/', '"')
        alt, pos = text.extract(page, 'alt="', '"', pos)
        return f"{self.root}/img/{url}", alt

    def metadata(self, page):
        pos = page.find('<img src="/img/')
        width = text.extract(page, 'width="', '"', pos)[0]
        height = text.extract(page, 'height="', '"', pos)[0]

        return {
            "width" : text.parse_int(width),
            "height": text.parse_int(height),
        }


class ImgpvImageExtractor(ImagehostImageExtractor):
    """Extractor for imgpv.com images"""
    category = "imgpv"
    root = "https://imgpv.com"
    pattern = (r"(?:https?://)?(?:www\.)?imgpv\.com"
               r"(/([a-z0-9]{10,})/[\S]+\.html)")
    example = "https://www.imgpv.com/a1b2c3d4f5g6/NAME.EXT.html"

    def get_info(self, page):
        url, pos = text.extract(page, 'id="img-preview" src="', '"')
        alt, pos = text.extract(page, 'alt="', '"', pos)
        return url, text.unescape(alt)

    def metadata(self, page):
        pos = page.find('class="upinfo">')
        date, pos = text.extract(page, '<b>', 'by', pos)
        user, pos = text.extract(page, '>', '<', pos)

        date = date.split()
        return {
            "date": self.parse_datetime_iso(f"{date[0][:10]} {date[1]}"),
            "user": text.unescape(user),
        }



# -*- coding: utf-8 -*-

# Copyright 2017-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Command line option parsing"""

import argparse
import logging
import os.path
import sys
from . import job, util, version


class ConfigAction(argparse.Action):
    """Set argparse results as config values"""
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.options.append(((), self.dest, values))


class ConfigConstAction(argparse.Action):
    """Set argparse const values as config values"""
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.options.append(((), self.dest, self.const))


class AppendCommandAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest, None) or []
        val = self.const.copy()
        val["command"] = values
        items.append(val)
        setattr(namespace, self.dest, items)


class DeprecatedConfigConstAction(argparse.Action):
    """Set argparse const values as config values + deprecation warning"""
    def __call__(self, parser, namespace, values, option_string=None):
        sys.stderr.write(
            f"Warning: {'/'.join(self.option_strings)} is deprecated. "
            f"Use {self.choices} instead.\n")
        namespace.options.append(((), self.dest, self.const))


class ConfigParseAction(argparse.Action):
    """Parse KEY=VALUE config options"""
    def __call__(self, parser, namespace, values, option_string=None):
        key, value = _parse_option(values)
        key = key.split(".")  # splitting an empty string becomes [""]
        namespace.options.append((key[:-1], key[-1], value))


class PPParseAction(argparse.Action):
    """Parse KEY=VALUE post processor options"""
    def __call__(self, parser, namespace, values, option_string=None):
        key, value = _parse_option(values)
        namespace.options_pp[key] = value


class InputfileAction(argparse.Action):
    """Collect input files"""
    def __call__(self, parser, namespace, value, option_string=None):
        namespace.input_files.append((value, self.const))


class MtimeAction(argparse.Action):
    """Configure mtime post processors"""
    def __call__(self, parser, namespace, value, option_string=None):
        namespace.postprocessors.append({
            "name": "mtime",
            "value": f"{{{self.const or value}}}",
        })


class RenameAction(argparse.Action):
    """Configure rename post processors"""
    def __call__(self, parser, namespace, value, option_string=None):
        if self.const:
            namespace.postprocessors.append({
                "name": "rename",
                "to"  : value,
            })
        else:
            namespace.postprocessors.append({
                "name": "rename",
                "from": value,
            })


class UgoiraAction(argparse.Action):
    """Configure ugoira post processors"""
    def __call__(self, parser, namespace, value, option_string=None):
        if self.const:
            value = self.const
        else:
            value = value.strip().lower()

        if value in ("webm", "vp9"):
            pp = {
                "extension"        : "webm",
                "ffmpeg-args"      : ("-c:v", "libvpx-vp9",
                                      "-crf", "12",
                                      "-b:v", "0", "-an"),
            }
        elif value == "vp9-lossless":
            pp = {
                "extension"        : "webm",
                "ffmpeg-args"      : ("-c:v", "libvpx-vp9",
                                      "-lossless", "1",
                                      "-pix_fmt", "yuv420p", "-an"),
            }
        elif value == "vp8":
            pp = {
                "extension"        : "webm",
                "ffmpeg-args"      : ("-c:v", "libvpx",
                                      "-crf", "4",
                                      "-b:v", "5000k", "-an"),
            }
        elif value == "mp4":
            pp = {
                "extension"        : "mp4",
                "ffmpeg-args"      : ("-c:v", "libx264", "-an", "-b:v", "5M"),
                "libx264-prevent-odd": True,
            }
        elif value == "gif":
            pp = {
                "extension"        : "gif",
                "ffmpeg-args"      : ("-filter_complex", "[0:v] split [a][b];"
                                      "[a] palettegen [p];[b][p] paletteuse"),
                "repeat-last-frame": False,
            }
        elif value == "mkv" or value == "copy":
            pp = {
                "extension"        : "mkv",
                "ffmpeg-args"      : ("-c:v", "copy"),
                "repeat-last-frame": False,
            }
        elif value == "zip" or value == "archive":
            pp = {
                "mode"             : "archive",
            }
            namespace.options.append(((), "ugoira", "original"))
        else:
            parser.error(f"Unsupported Ugoira format '{value}'")

        pp["name"] = "ugoira"
        pp["whitelist"] = ("pixiv", "danbooru")

        namespace.options.append((("extractor",), "ugoira", True))
        namespace.postprocessors.append(pp)


class PrintAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        if self.const:
            if self.const == "-":
                namespace.options.append(((), "skip", False))
                namespace.options.append(((), "download", False))
                namespace.options.append((("output",), "mode", False))
            filename = "-"
            base = None
            mode = "w"
        else:
            if self.const is None:
                namespace.options.append(((), "skip", False))
                namespace.options.append(((), "download", False))
            value, path = value
            base, filename = os.path.split(path)
            mode = "a"

        event, sep, format_string = value.partition(":")
        if not sep:
            format_string = event
            event = ("prepare",)
        else:
            event = event.strip().lower()
            if event not in {"init", "file", "after", "skip", "error",
                             "prepare", "prepare-after", "post", "post-after",
                             "finalize", "finalize-success", "finalize-error"}:
                format_string = value
                event = ("prepare",)

        if not format_string:
            return

        if format_string.startswith("\\f"):
            format_string = "\f" + format_string[2:]

        if format_string[0] == "\f":
            if format_string[1] == "F" and format_string[-1] != "\n":
                format_string += "\n"
        elif "{" not in format_string and " " not in format_string:
            format_string = f"{{{format_string}}}\n"
        elif format_string[-1] != "\n":
            format_string += "\n"

        namespace.postprocessors.append({
            "name"          : "metadata",
            "event"         : event,
            "filename"      : filename,
            "base-directory": base or ".",
            "content-format": format_string,
            "open"          : mode,
        })


class Formatter(argparse.HelpFormatter):
    """Custom HelpFormatter class to customize help output"""
    def __init__(self, prog):
        argparse.HelpFormatter.__init__(self, prog, max_help_position=30)

    def _format_action_invocation(self, action):
        opts = action.option_strings
        if action.metavar:
            opts = opts.copy()
            opts[-1] = f"{opts[-1]} {action.metavar}"
        return ", ".join(opts)

    def _format_usage(self, usage, actions, groups, prefix):
        return f"Usage: {self._prog} [OPTIONS] URL [URL...]\n"


def _parse_option(opt):
    key, _, value = opt.partition("=")
    try:
        value = util.json_loads(value)
    except ValueError:
        pass
    return key, value


def build_parser():
    """Build and configure an ArgumentParser object"""
    parser = argparse.ArgumentParser(
        formatter_class=Formatter,
        add_help=False,
    )

    general = parser.add_argument_group("General Options")
    general.add_argument(
        "-h", "--help",
        action="help",
        help="Print this help message and exit",
    )
    general.add_argument(
        "--version",
        action="version", version=version.__version__,
        help="Print program version and exit",
    )
    general.add_argument(
        "-f", "--filename",
        dest="filename", metavar="FORMAT",
        help=("Filename format string for downloaded files "
              "('/O' for \"original\" filenames)"),
    )
    general.add_argument(
        "-d", "--destination",
        dest="base-directory", metavar="PATH", action=ConfigAction,
        help="Target location for file downloads",
    )
    general.add_argument(
        "-D", "--directory",
        dest="directory", metavar="PATH",
        help="Exact location for file downloads",
    )
    general.add_argument(
        "-X", "--extractors",
        dest="extractor_sources", metavar="PATH", action="append",
        help="Load external extractors from PATH",
    )
    general.add_argument(
        "-a", "--user-agent",
        dest="user-agent", metavar="UA", action=ConfigAction,
        help="User-Agent request header",
    )
    general.add_argument(
        "--clear-cache",
        dest="clear_cache", metavar="MODULE",
        help="Delete cached login sessions, cookies, etc. for MODULE "
             "(ALL to delete everything)",
    )
    general.add_argument(
        "--compat",
        dest="category-map", nargs=0, action=ConfigConstAction, const="compat",
        help="Restore legacy 'category' names",
    )

    update = parser.add_argument_group("Update Options")
    if util.EXECUTABLE:
        update.add_argument(
            "-U", "--update",
            dest="update", action="store_const", const="latest",
            help="Update to the latest version",
        )
        update.add_argument(
            "--update-to",
            dest="update", metavar="CHANNEL[@TAG]",
            help=("Switch to a dfferent release channel (stable or dev) "
                  "or upgrade/downgrade to a specific version"),
        )
        update.add_argument(
            "--update-check",
            dest="update", action="store_const", const="check",
            help="Check if a newer version is available",
        )
    else:
        update.add_argument(
            "-U", "--update-check",
            dest="update", action="store_const", const="check",
            help="Check if a newer version is available",
        )

    input = parser.add_argument_group("Input Options")
    input.add_argument(
        "urls",
        metavar="URL", nargs="*",
        help=argparse.SUPPRESS,
    )
    input.add_argument(
        "-i", "--input-file",
        dest="input_files", metavar="FILE", action=InputfileAction, const=None,
        default=[],
        help=("Download URLs found in FILE ('-' for stdin). "
              "More than one --input-file can be specified"),
    )
    input.add_argument(
        "-I", "--input-file-comment",
        dest="input_files", metavar="FILE", action=InputfileAction, const="c",
        help=("Download URLs found in FILE. "
              "Comment them out after they were downloaded successfully."),
    )
    input.add_argument(
        "-x", "--input-file-delete",
        dest="input_files", metavar="FILE", action=InputfileAction, const="d",
        help=("Download URLs found in FILE. "
              "Delete them after they were downloaded successfully."),
    )
    input.add_argument(
        "--no-input",
        dest="input", nargs=0, action=ConfigConstAction, const=False,
        help="Do not prompt for passwords/tokens",
    )

    output = parser.add_argument_group("Output Options")
    output.add_argument(
        "-q", "--quiet",
        dest="loglevel", default=logging.INFO,
        action="store_const", const=logging.ERROR,
        help="Activate quiet mode",
    )
    output.add_argument(
        "-w", "--warning",
        dest="loglevel",
        action="store_const", const=logging.WARNING,
        help="Print only warnings and errors",
    )
    output.add_argument(
        "-v", "--verbose",
        dest="loglevel",
        action="store_const", const=logging.DEBUG,
        help="Print various debugging information",
    )
    output.add_argument(
        "-g", "--get-urls",
        dest="list_urls", action="count",
        help="Print URLs instead of downloading",
    )
    output.add_argument(
        "-G", "--resolve-urls",
        dest="list_urls", action="store_const", const=128,
        help="Print URLs instead of downloading; resolve intermediary URLs",
    )
    output.add_argument(
        "-j", "--dump-json",
        dest="dump_json", action="count",
        help="Print JSON information",
    )
    output.add_argument(
        "-J", "--resolve-json",
        dest="dump_json", action="store_const", const=128,
        help="Print JSON information; resolve intermediary URLs",
    )
    output.add_argument(
        "-s", "--simulate",
        dest="jobtype", action="store_const", const=job.SimulationJob,
        help="Simulate data extraction; do not download anything",
    )
    output.add_argument(
        "-E", "--extractor-info",
        dest="jobtype", action="store_const", const=job.InfoJob,
        help="Print extractor defaults and settings",
    )
    output.add_argument(
        "-K", "--list-keywords",
        dest="jobtype", action="store_const", const=job.KeywordJob,
        help=("Print a list of available keywords and example values "
              "for the given URLs"),
    )
    output.add_argument(
        "-e", "--error-file",
        dest="errorfile", metavar="FILE", action=ConfigAction,
        help="Add input URLs which returned an error to FILE",
    )
    output.add_argument(
        "-N", "--print",
        dest="postprocessors", metavar="[EVENT:]FORMAT",
        action=PrintAction, const="-", default=[],
        help=("Write FORMAT during EVENT (default 'prepare') to standard "
              "output instead of downloading files. "
              "Can be used multiple times. "
              "Examples: 'id' or 'post:{md5[:8]}'"),
    )
    output.add_argument(
        "--Print",
        dest="postprocessors", metavar="[EVENT:]FORMAT",
        action=PrintAction, const="+",
        help="Like --print, but downloads files as well",
    )
    output.add_argument(
        "--print-to-file",
        dest="postprocessors", metavar="[EVENT:]FORMAT FILE",
        action=PrintAction, const=None, nargs=2,
        help=("Append FORMAT during EVENT to FILE instead of downloading "
              "files. Can be used multiple times"),
    )
    output.add_argument(
        "--Print-to-file",
        dest="postprocessors", metavar="[EVENT:]FORMAT FILE",
        action=PrintAction, const=False, nargs=2,
        help="Like --print-to-file, but downloads files as well",
    )
    output.add_argument(
        "--list-modules",
        dest="list_modules", action="store_true",
        help="Print a list of available extractor modules",
    )
    output.add_argument(
        "--list-extractors",
        dest="list_extractors", metavar="[CATEGORIES]", nargs="*",
        help=("Print a list of extractor classes "
              "with description, (sub)category and example URL"),
    )
    output.add_argument(
        "--write-log",
        dest="logfile", metavar="FILE", action=ConfigAction,
        help="Write logging output to FILE",
    )
    output.add_argument(
        "--write-unsupported",
        dest="unsupportedfile", metavar="FILE", action=ConfigAction,
        help=("Write URLs, which get emitted by other extractors but cannot "
              "be handled, to FILE"),
    )
    output.add_argument(
        "--write-pages",
        dest="write-pages", nargs=0, action=ConfigConstAction, const=True,
        help=("Write downloaded intermediary pages to files "
              "in the current directory to debug problems"),
    )
    output.add_argument(
        "--print-traffic",
        dest="print_traffic", action="store_true",
        help="Display sent and read HTTP traffic",
    )
    output.add_argument(
        "--no-colors",
        dest="colors", action="store_false",
        help="Do not emit ANSI color codes in output",
    )

    networking = parser.add_argument_group("Networking Options")
    networking.add_argument(
        "-R", "--retries",
        dest="retries", metavar="N", type=int, action=ConfigAction,
        help=("Maximum number of retries for failed HTTP requests "
              "or -1 for infinite retries (default: 4)"),
    )
    networking.add_argument(
        "--http-timeout",
        dest="timeout", metavar="SECONDS", type=float, action=ConfigAction,
        help="Timeout for HTTP connections (default: 30.0)",
    )
    networking.add_argument(
        "--proxy",
        dest="proxy", metavar="URL", action=ConfigAction,
        help="Use the specified proxy",
    )
    networking.add_argument(
        "--source-address",
        dest="source-address", metavar="IP", action=ConfigAction,
        help="Client-side IP address to bind to",
    )
    networking.add_argument(
        "-4", "--force-ipv4",
        dest="source-address", nargs=0, action=ConfigConstAction,
        const="0.0.0.0",
        help="Make all connections via IPv4",
    )
    networking.add_argument(
        "-6", "--force-ipv6",
        dest="source-address", nargs=0, action=ConfigConstAction, const="::",
        help="Make all connections via IPv6",
    )
    networking.add_argument(
        "--no-check-certificate",
        dest="verify", nargs=0, action=ConfigConstAction, const=False,
        help="Disable HTTPS certificate validation",
    )

    downloader = parser.add_argument_group("Downloader Options")
    downloader.add_argument(
        "-r", "--limit-rate",
        dest="rate", metavar="RATE", action=ConfigAction,
        help="Maximum download rate (e.g. 500k, 2.5M, or 800k-2M)",
    )
    downloader.add_argument(
        "--chunk-size",
        dest="chunk-size", metavar="SIZE", action=ConfigAction,
        help="Size of in-memory data chunks (default: 32k)",
    )
    downloader.add_argument(
        "--sleep",
        dest="sleep", metavar="SECONDS", action=ConfigAction,
        help=("Number of seconds to wait before each download. "
              "This can be either a constant value or a range "
              "(e.g. 2.7 or 2.0-3.5)"),
    )
    downloader.add_argument(
        "--sleep-skip",
        dest="sleep-skip", metavar="SECONDS", action=ConfigAction,
        help=("Number of seconds to wait after skipping a file download"),
    )
    downloader.add_argument(
        "--sleep-request",
        dest="sleep-request", metavar="SECONDS", action=ConfigAction,
        help=("Number of seconds to wait between HTTP requests "
              "during data extraction"),
    )
    downloader.add_argument(
        "--sleep-429",
        dest="sleep-429", metavar="SECONDS", action=ConfigAction,
        help=("Number of seconds to wait when receiving a "
              "'429 Too Many Requests' response"),
    )
    downloader.add_argument(
        "--sleep-extractor",
        dest="sleep-extractor", metavar="SECONDS", action=ConfigAction,
        help=("Number of seconds to wait before starting data extraction "
              "for an input URL"),
    )
    downloader.add_argument(
        "--no-part",
        dest="part", nargs=0, action=ConfigConstAction, const=False,
        help="Do not use .part files",
    )
    downloader.add_argument(
        "--no-skip",
        dest="skip", nargs=0, action=ConfigConstAction, const=False,
        help="Do not skip downloads; overwrite existing files",
    )
    downloader.add_argument(
        "--no-mtime",
        dest="mtime", nargs=0, action=ConfigConstAction, const=False,
        help=("Do not set file modification times according to "
              "Last-Modified HTTP response headers")
    )
    downloader.add_argument(
        "--no-download",
        dest="download", nargs=0, action=ConfigConstAction, const=False,
        help=("Do not download any files")
    )

    configuration = parser.add_argument_group("Configuration Options")
    configuration.add_argument(
        "-o", "--option",
        dest="options", metavar="KEY=VALUE",
        action=ConfigParseAction, default=[],
        help=("Additional options. "
              "Example: -o browser=firefox")   ,
    )
    configuration.add_argument(
        "-c", "--config",
        dest="configs_json", metavar="FILE", action="append",
        help="Additional configuration files",
    )
    configuration.add_argument(
        "--config-yaml",
        dest="configs_yaml", metavar="FILE", action="append",
        help="Additional configuration files in YAML format",
    )
    configuration.add_argument(
        "--config-toml",
        dest="configs_toml", metavar="FILE", action="append",
        help="Additional configuration files in TOML format",
    )
    configuration.add_argument(
        "--config-create",
        dest="config", action="store_const", const="init",
        help="Create a basic configuration file",
    )
    configuration.add_argument(
        "--config-status",
        dest="config", action="store_const", const="status",
        help="Show configuration file status",
    )
    configuration.add_argument(
        "--config-open",
        dest="config", action="store_const", const="open",
        help="Open configuration file in external application",
    )
    configuration.add_argument(
        "--config-ignore",
        dest="config_load", action="store_false",
        help="Do not read default configuration files",
    )
    configuration.add_argument(
        "--ignore-config",
        dest="config_load", action="store_false",
        help=argparse.SUPPRESS,
    )

    authentication = parser.add_argument_group("Authentication Options")
    authentication.add_argument(
        "-u", "--username",
        dest="username", metavar="USER", action=ConfigAction,
        help="Username to login with",
    )
    authentication.add_argument(
        "-p", "--password",
        dest="password", metavar="PASS", action=ConfigAction,
        help="Password belonging to the given username",
    )
    authentication.add_argument(
        "--netrc",
        dest="netrc", nargs=0, action=ConfigConstAction, const=True,
        help="Enable .netrc authentication data",
    )

    cookies = parser.add_argument_group("Cookie Options")
    cookies.add_argument(
        "-C", "--cookies",
        dest="cookies", metavar="FILE", action=ConfigAction,
        help="File to load additional cookies from",
    )
    cookies.add_argument(
        "--cookies-export",
        dest="cookies-update", metavar="FILE", action=ConfigAction,
        help="Export session cookies to FILE",
    )
    cookies.add_argument(
        "--cookies-from-browser",
        dest="cookies_from_browser",
        metavar="BROWSER[/DOMAIN][+KEYRING][:PROFILE][::CONTAINER]",
        help=("Name of the browser to load cookies from, with optional "
              "domain prefixed with '/', "
              "keyring name prefixed with '+', "
              "profile prefixed with ':', and "
              "container prefixed with '::' "
              "('none' for no container (default), 'all' for all containers)"),
    )

    selection = parser.add_argument_group("Selection Options")
    selection.add_argument(
        "-A", "--abort",
        dest="abort", metavar="N[:TARGET]",
        help=("Stop current extractor(s) "
              "after N consecutive file downloads were skipped. "
              "Specify a TARGET to set how many levels to ascend or "
              "to which subcategory to jump to. "
              "Examples: '-A 3', '-A 3:2', '-A 3:manga'"),
    )
    selection.add_argument(
        "-T", "--terminate",
        dest="terminate", metavar="N",
        help=("Stop current & parent extractors "
              "and proceed with the next input URL "
              "after N consecutive file downloads were skipped"),
    )
    selection.add_argument(
        "--filesize-min",
        dest="filesize-min", metavar="SIZE", action=ConfigAction,
        help="Do not download files smaller than SIZE (e.g. 500k or 2.5M)",
    )
    selection.add_argument(
        "--filesize-max",
        dest="filesize-max", metavar="SIZE", action=ConfigAction,
        help="Do not download files larger than SIZE (e.g. 500k or 2.5M)",
    )
    selection.add_argument(
        "--download-archive",
        dest="archive", metavar="FILE", action=ConfigAction,
        help=("Record successfully downloaded files in FILE and "
              "skip downloading any file already in it"),
    )
    selection.add_argument(
        "--range",
        dest="image-range", metavar="RANGE", action=ConfigAction,
        help=("Index range(s) specifying which files to download. "
              "These can be either a constant value, range, or slice "
              "(e.g. '5', '8-20', or '1:24:3')"),
    )
    selection.add_argument(
        "--post-range",
        dest="post-range", metavar="RANGE", action=ConfigAction,
        help=("Like '--range', but for posts"),
    )
    selection.add_argument(
        "--chapter-range",
        dest="chapter-range", metavar="RANGE", action=ConfigAction,
        help=("Like '--range', but for child extractors handling "
              "manga chapters, external URLs, etc."),
    )
    selection.add_argument(
        "--filter",
        dest="image-filter", metavar="EXPR", action=ConfigAction,
        help=("Python expression controlling which files to download. "
              "Files for which the expression evaluates to False are ignored. "
              "Available keys are the filename-specific ones listed by '-K'. "
              "Example: --filter \"image_width >= 1000 and "
              "rating in ('s', 'q')\""),
    )
    selection.add_argument(
        "--post-filter",
        dest="post-filter", metavar="EXPR", action=ConfigAction,
        help=("Like '--filter', but for posts"),
    )
    selection.add_argument(
        "--chapter-filter",
        dest="chapter-filter", metavar="EXPR", action=ConfigAction,
        help=("Like '--filter', but for child extractors handling "
              "manga chapters, external URLs, etc."),
    )

    infojson = {
        "name"    : "metadata",
        "event"   : "init",
        "filename": "info.json",
    }
    postprocessor = parser.add_argument_group("Post-processing Options")
    postprocessor.add_argument(
        "-P", "--postprocessor",
        dest="postprocessors", metavar="NAME", action="append",
        help="Activate the specified post processor",
    )
    postprocessor.add_argument(
        "--no-postprocessors",
        dest="postprocess", nargs=0, action=ConfigConstAction, const=False,
        help=("Do not run any post processors")
    )
    postprocessor.add_argument(
        "-O", "--postprocessor-option",
        dest="options_pp", metavar="KEY=VALUE",
        action=PPParseAction, default={},
        help="Additional post processor options",
    )
    postprocessor.add_argument(
        "--write-metadata",
        dest="postprocessors",
        action="append_const", const="metadata",
        help="Write metadata to separate JSON files",
    )
    postprocessor.add_argument(
        "--write-info-json",
        dest="postprocessors",
        action="append_const", const=infojson,
        help="Write gallery metadata to a info.json file",
    )
    postprocessor.add_argument(
        "--write-infojson",
        dest="postprocessors",
        action="append_const", const=infojson,
        help=argparse.SUPPRESS,
    )
    postprocessor.add_argument(
        "--write-tags",
        dest="postprocessors",
        action="append_const", const={"name": "metadata", "mode": "tags"},
        help="Write image tags to separate text files",
    )
    postprocessor.add_argument(
        "--zip",
        dest="postprocessors",
        action="append_const", const="zip",
        help="Store downloaded files in a ZIP archive",
    )
    postprocessor.add_argument(
        "--cbz",
        dest="postprocessors",
        action="append_const", const={
            "name"     : "zip",
            "extension": "cbz",
        },
        help="Store downloaded files in a CBZ archive",
    )
    postprocessor.add_argument(
        "--mtime",
        dest="postprocessors", metavar="NAME", action=MtimeAction,
        help=("Set file modification times according to metadata "
              "selected by NAME. Examples: 'date' or 'status[date]'"),
    )
    postprocessor.add_argument(
        "--mtime-from-date",
        dest="postprocessors", nargs=0, action=MtimeAction,
        const="date|status[date]",
        help=argparse.SUPPRESS,
    )
    postprocessor.add_argument(
        "--rename",
        dest="postprocessors", metavar="FORMAT", action=RenameAction, const=0,
        help=("Rename previously downloaded files from FORMAT "
              "to the current filename format"),
    )
    postprocessor.add_argument(
        "--rename-to",
        dest="postprocessors", metavar="FORMAT", action=RenameAction, const=1,
        help=("Rename previously downloaded files from the current filename "
              "format to FORMAT"),
    )
    postprocessor.add_argument(
        "--ugoira",
        dest="postprocessors", metavar="FMT", action=UgoiraAction,
        help=("Convert Pixiv Ugoira to FMT using FFmpeg. "
              "Supported formats are 'webm', 'mp4', 'gif', "
              "'vp8', 'vp9', 'vp9-lossless', 'copy', 'zip'."),
    )
    postprocessor.add_argument(
        "--ugoira-conv",
        dest="postprocessors", nargs=0, action=UgoiraAction, const="vp8",
        help=argparse.SUPPRESS,
    )
    postprocessor.add_argument(
        "--ugoira-conv-lossless",
        dest="postprocessors", nargs=0, action=UgoiraAction,
        const="vp9-lossless",
        help=argparse.SUPPRESS,
    )
    postprocessor.add_argument(
        "--ugoira-conv-copy",
        dest="postprocessors", nargs=0, action=UgoiraAction, const="copy",
        help=argparse.SUPPRESS,
    )
    postprocessor.add_argument(
        "--exec",
        dest="postprocessors", metavar="CMD",
        action=AppendCommandAction, const={"name": "exec"},
        help=("Execute CMD for each downloaded file. "
              "Supported replacement fields are "
              "{} or {_path}, {_temppath}, {_directory}, {_filename}. "
              "On Windows, use {_path_unc} or {_directory_unc} for UNC paths. "
              "Example: --exec \"convert {} {}.png && rm {}\""),
    )
    postprocessor.add_argument(
        "--exec-after",
        dest="postprocessors", metavar="CMD",
        action=AppendCommandAction, const={
            "name": "exec", "event": "finalize"},
        help=("Execute CMD after all files were downloaded. "
              "Example: --exec-after \"cd {_directory} "
              "&& convert * ../doc.pdf\""),
    )

    try:
        # restore normal behavior when adding '-4' or '-6' as arguments
        parser._has_negative_number_optionals.clear()
    except Exception:
        pass

    return parser



# -*- coding: utf-8 -*-

# Copyright 2024-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://realbooru.com/"""

from . import booru
from .. import text, util
import collections

BASE_PATTERN = r"(?:https?://)?realbooru\.com"


class RealbooruExtractor(booru.BooruExtractor):
    basecategory = "booru"
    category = "realbooru"
    root = "https://realbooru.com"

    def _parse_post(self, post_id):
        url = f"{self.root}/index.php?page=post&s=view&id={post_id}"
        page = self.request(url).text
        extr = text.extract_from(page)
        rating = extr('name="rating" content="', '"')
        extr('class="container"', '>')

        post = {
            "id"        : post_id,
            "rating"    : "e" if rating == "adult" else (rating or "?")[0],
            "file_url"  : (s := extr('src="', '"')),
            "_fallback" : (extr('src="', '"'),) if s.endswith(".mp4") else (),
            "created_at": extr(">Posted at ", " by "),
            "uploader"  : extr(">", "<"),
            "score"     : extr('">', "<"),
            "tags"      : extr('<br />', "</div>"),
            "title"     : extr('id="title" style="width: 100%;" value="', '"'),
            "source"    : extr('d="source" style="width: 100%;" value="', '"'),
        }

        tags_container = post["tags"]
        tags = []
        tags_categories = collections.defaultdict(list)
        pattern = text.re(r'<a class="(?:tag-type-)?([^"]+).*?;tags=([^"&]+)')
        for tag_type, tag_name in pattern.findall(tags_container):
            tag = text.unescape(text.unquote(tag_name))
            tags.append(tag)
            tags_categories[tag_type].append(tag)
        for key, value in tags_categories.items():
            post["tags_" + key] = ", ".join(value)
        tags.sort()

        post["tags"] = ", ".join(tags)
        post["md5"] = post["file_url"].rpartition("/")[2].partition(".")[0]
        return post

    def skip(self, num):
        self.page_start += num
        return num

    def _prepare(self, post):
        post["date"] = self.parse_datetime(post["created_at"], "%b, %d %Y")

    def _pagination(self, params, begin, end):
        url = self.root + "/index.php"
        params["pid"] = self.page_start

        while True:
            page = self.request(url, params=params).text

            cnt = 0
            for post_id in text.extract_iter(page, begin, end):
                cnt += 1
                yield self._parse_post(post_id)

            if cnt < self.per_page:
                return
            params["pid"] += self.per_page


class RealbooruTagExtractor(RealbooruExtractor):
    subcategory = "tag"
    directory_fmt = ("{category}", "{search_tags}")
    archive_fmt = "t_{search_tags}_{id}"
    per_page = 42
    pattern = BASE_PATTERN + r"/index\.php\?page=post&s=list&tags=([^&#]*)"
    example = "https://realbooru.com/index.php?page=post&s=list&tags=TAG"

    def metadata(self):
        self.tags = text.unquote(self.groups[0].replace("+", " "))
        return {"search_tags": self.tags}

    def posts(self):
        return self._pagination({
            "page": "post",
            "s"   : "list",
            "tags": self.tags,
        }, '<a id="p', '"')


class RealbooruFavoriteExtractor(RealbooruExtractor):
    subcategory = "favorite"
    directory_fmt = ("{category}", "favorites", "{favorite_id}")
    archive_fmt = "f_{favorite_id}_{id}"
    per_page = 50
    pattern = BASE_PATTERN + r"/index\.php\?page=favorites&s=view&id=(\d+)"
    example = "https://realbooru.com/index.php?page=favorites&s=view&id=12345"

    def metadata(self):
        return {"favorite_id": text.parse_int(self.groups[0])}

    def posts(self):
        return self._pagination({
            "page": "favorites",
            "s"   : "view",
            "id"  : self.groups[0],
        }, '" id="p', '"')


class RealbooruPoolExtractor(RealbooruExtractor):
    subcategory = "pool"
    directory_fmt = ("{category}", "pool", "{pool} {pool_name}")
    archive_fmt = "p_{pool}_{id}"
    pattern = BASE_PATTERN + r"/index\.php\?page=pool&s=show&id=(\d+)"
    example = "https://realbooru.com/index.php?page=pool&s=show&id=12345"

    def metadata(self):
        pool_id = self.groups[0]
        url = f"{self.root}/index.php?page=pool&s=show&id={pool_id}"
        page = self.request(url).text

        name, pos = text.extract(page, "<h4>Pool: ", "</h4>")
        self.post_ids = text.extract_iter(
            page, 'class="thumb" id="p', '"', pos)

        return {
            "pool": text.parse_int(pool_id),
            "pool_name": text.unescape(name),
        }

    def posts(self):
        return map(
            self._parse_post,
            util.advance(self.post_ids, self.page_start)
        )


class RealbooruPostExtractor(RealbooruExtractor):
    subcategory = "post"
    archive_fmt = "{id}"
    pattern = BASE_PATTERN + r"/index\.php\?page=post&s=view&id=(\d+)"
    example = "https://realbooru.com/index.php?page=post&s=view&id=12345"

    def posts(self):
        return (self._parse_post(self.groups[0]),)



# -*- coding: utf-8 -*-

# Copyright 2016-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://hentai2read.com/"""

from .common import ChapterExtractor, MangaExtractor
from .. import text, util


class Hentai2readBase():
    """Base class for hentai2read extractors"""
    category = "hentai2read"
    root = "https://hentai2read.com"


class Hentai2readChapterExtractor(Hentai2readBase, ChapterExtractor):
    """Extractor for a single manga chapter from hentai2read.com"""
    archive_fmt = "{chapter_id}_{page}"
    pattern = r"(?:https?://)?(?:www\.)?hentai2read\.com(/[^/?#]+/([^/?#]+))"
    example = "https://hentai2read.com/TITLE/1/"

    def metadata(self, page):
        title, pos = text.extract(page, "<title>", "</title>")
        manga_id, pos = text.extract(page, 'data-mid="', '"', pos)
        chapter_id, pos = text.extract(page, 'data-cid="', '"', pos)
        chapter, sep, minor = self.groups[1].partition(".")

        match = text.re(
            r"Reading (.+) \(([^)]+)\) Hentai(?: by (.*))? - "
            r"([^:]+): (.+) . Page 1 ").match(title)
        if match:
            manga, type, author, _, title = match.groups()
        else:
            self.log.warning("Failed to extract 'manga', 'type', 'author', "
                             "and 'title' metadata")
            manga = type = author = title = ""

        return {
            "manga": manga,
            "manga_id": text.parse_int(manga_id),
            "chapter": text.parse_int(chapter),
            "chapter_minor": sep + minor,
            "chapter_id": text.parse_int(chapter_id),
            "type": type,
            "author": author,
            "title": title,
            "lang": "en",
            "language": "English",
        }

    def images(self, page):
        images = text.extract(page, "'images' : ", ",\n")[0]
        return [
            ("https://hentaicdn.com/hentai" + part, None)
            for part in util.json_loads(images)
        ]


class Hentai2readMangaExtractor(Hentai2readBase, MangaExtractor):
    """Extractor for hmanga from hentai2read.com"""
    chapterclass = Hentai2readChapterExtractor
    pattern = r"(?:https?://)?(?:www\.)?hentai2read\.com(/[^/?#]+)/?$"
    example = "https://hentai2read.com/TITLE/"

    def chapters(self, page):
        results = []

        pos = page.find('itemscope itemtype="http://schema.org/Book') + 1
        manga, pos = text.extract(
            page, '<span itemprop="name">', '</span>', pos)
        mtype, pos = text.extract(
            page, '<small class="text-danger">[', ']</small>', pos)
        manga_id = text.parse_int(text.extract(
            page, 'data-mid="', '"', pos)[0])

        while True:
            chapter_id, pos = text.extract(page, ' data-cid="', '"', pos)
            if not chapter_id:
                return results
            _  , pos = text.extract(page, ' href="', '"', pos)
            url, pos = text.extract(page, ' href="', '"', pos)

            chapter, pos = text.extract(page, '>', '<', pos)
            chapter, _, title = text.unescape(chapter).strip().partition(" - ")
            chapter, sep, minor = chapter.partition(".")

            results.append((url, {
                "manga": manga,
                "manga_id": manga_id,
                "chapter": text.parse_int(chapter),
                "chapter_minor": sep + minor,
                "chapter_id": text.parse_int(chapter_id),
                "type": mtype,
                "title": title,
                "lang": "en",
                "language": "English",
            }))



# -*- coding: utf-8 -*-

# Copyright 2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://fikfap.com/"""

from .common import Extractor, Message
from .. import text, exception

BASE_PATTERN = r"(?:https?://)?(?:www\.)?fikfap\.com"


class FikfapExtractor(Extractor):
    """Base class for fikfap extractors"""
    category = "fikfap"
    root = "https://fikfap.com"
    root_api = "https://api.fikfap.com"
    directory_fmt = ("{category}", "{author[username]}")
    filename_fmt = "{postId} {label[:240]}.{extension}"
    archive_fmt = "{postId}"

    def items(self):
        headers = {
            "Referer"       : self.root + "/",
            "Origin"        : self.root,
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "cross-site",
        }

        for post in self.posts():
            if url := post.get("videoFileOriginalUrl"):
                post["extension"] = text.ext_from_url(url)
            elif url := post.get("videoStreamUrl"):
                url = "ytdl:" + url
                post["extension"] = "mp4"
                post["_ytdl_manifest"] = "hls"
                post["_ytdl_manifest_headers"] = headers
            else:
                self.log.warning("%s: No video available", post["postId"])
                continue

            post["date"] = self.parse_datetime_iso(post["createdAt"])
            post["date_updated"] = self.parse_datetime_iso(post["updatedAt"])
            post["tags"] = [t["label"] for t in post["hashtags"]]
            post["filename"] = post["label"]

            yield Message.Directory, "", post
            yield Message.Url, url, post

    def request_api(self, url, params):
        return self.request_json(url, params=params, headers={
            "Referer"       : self.root + "/",
            "Authorization-Anonymous": "2527cc30-c3c5-41be-b8bb-104b6ea7a206",
            "IsLoggedIn"    : "false",
            "IsPWA"         : "false",
            "Origin"        : self.root,
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-site",
        })


class FikfapPostExtractor(FikfapExtractor):
    subcategory = "post"
    pattern = BASE_PATTERN + r"/(?:user/[^/?#]+/)?post/(\d+)"
    example = "https://fikfap.com/user/USER/post/12345"

    def posts(self):
        pid = self.groups[0]

        url = f"{self.root_api}/posts/{pid}"
        post = self.request_api(url, None)

        if post["postId"] == int(pid):
            return (post,)
        raise exception.NotFoundError("post")


class FikfapUserExtractor(FikfapExtractor):
    subcategory = "user"
    pattern = BASE_PATTERN + r"/user/([^/?#]+)"
    example = "https://fikfap.com/user/USER"

    def posts(self):
        user = self.groups[0]

        url = f"{self.root_api}/profile/username/{user}/posts"
        params = {"amount": "21"}

        while True:
            data = self.request_api(url, params)

            yield from data

            if len(data) < 21:
                return
            params["afterId"] = data[-1]["postId"]


class FikfapHashtagExtractor(FikfapExtractor):
    subcategory = "hashtag"
    directory_fmt = ("{category}", "{hashtag}")
    pattern = BASE_PATTERN + r"/hash/([^/?#]+)"
    example = "https://fikfap.com/hash/HASH"

    def posts(self):
        self.kwdict["hashtag"] = hashtag = self.groups[0]

        url = f"{self.root_api}/hashtags/label/{hashtag}/posts"
        params = {"amount": "21"}

        while True:
            data = self.request_api(url, params)

            yield from data

            if len(data) < 21:
                return
            params["afterId"] = data[-1]["postId"]



# -*- coding: utf-8 -*-

# Copyright 2018-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Post-processing modules"""

modules = {
    "actions",
    "classify",
    "compare",
    "directory",
    "exec",
    "hash",
    "metadata",
    "mtime",
    "python",
    "rename",
    "ugoira",
    "zip",
}


def find(name):
    """Return a postprocessor class with the given name"""
    try:
        return _cache[name]
    except KeyError:
        pass

    cls = None
    if name in modules:  # prevent unwanted imports
        try:
            module = __import__(name, globals(), None, None, 1)
        except ImportError:
            pass
        else:
            cls = module.__postprocessor__
    _cache[name] = cls
    return cls


# --------------------------------------------------------------------
# internals

_cache = {}



# -*- coding: utf-8 -*-

# Copyright 2019-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://www.weibo.com/"""

from .common import Extractor, Message, Dispatch
from .. import text, util, exception
from ..cache import cache
import random

BASE_PATTERN = r"(?:https?://)?(?:www\.|m\.)?weibo\.c(?:om|n)"
USER_PATTERN = BASE_PATTERN + r"/(?:(u|n|p(?:rofile)?)/)?([^/?#]+)(?:/home)?"


class WeiboExtractor(Extractor):
    category = "weibo"
    directory_fmt = ("{category}", "{user[screen_name]}")
    filename_fmt = "{status[id]}_{num:>02}.{extension}"
    archive_fmt = "{status[id]}_{num}"
    cookies_domain = ".weibo.com"
    cookies_names = ("SUB", "SUBP")
    root = "https://weibo.com"
    request_interval = (1.0, 2.0)

    def __init__(self, match):
        Extractor.__init__(self, match)
        self._prefix = match[1]
        self.user = match[2]

    def _init(self):
        self.livephoto = self.config("livephoto", True)
        self.retweets = self.config("retweets", False)
        self.longtext = self.config("text", False)
        self.videos = self.config("videos", True)
        self.movies = self.config("movies", False)
        self.gifs = self.config("gifs", True)
        self.gifs_video = (self.gifs == "video")

        cookies = _cookie_cache()
        if cookies is None:
            self.logged_in = self.cookies_check(
                self.cookies_names, self.cookies_domain)
            return

        domain = self.cookies_domain
        cookies = {c.name: c for c in cookies if c.domain == domain}
        for cookie in self.cookies:
            if cookie.domain == domain and cookie.name in cookies:
                del cookies[cookie.name]
                if not cookies:
                    self.logged_in = True
                    return

        self.logged_in = False
        for cookie in cookies.values():
            self.cookies.set_cookie(cookie)

    def request(self, url, **kwargs):
        response = Extractor.request(self, url, **kwargs)

        if response.history:
            if "login.sina.com" in response.url:
                raise exception.AbortExtraction(
                    f"HTTP redirect to login page "
                    f"({response.url.partition('?')[0]})")
            if "passport.weibo.com" in response.url:
                self._sina_visitor_system(response)
                response = Extractor.request(self, url, **kwargs)

        return response

    def items(self):
        original_retweets = (self.retweets == "original")

        for status in self.statuses():

            if "ori_mid" in status and not self.retweets:
                self.log.debug("Skipping %s (快转 retweet)", status["id"])
                continue

            if "retweeted_status" in status:
                if not self.retweets:
                    self.log.debug("Skipping %s (retweet)", status["id"])
                    continue

                # videos of the original post are in status
                # images of the original post are in status["retweeted_status"]
                files = []
                self._extract_status(status, files)
                self._extract_status(status["retweeted_status"], files)

                if original_retweets:
                    status = status["retweeted_status"]
            else:
                files = []
                self._extract_status(status, files)

            if self.longtext and status.get("isLongText") and \
                    status["text"].endswith('class="expand">展开</span>'):
                status = self._status_by_id(status["id"])

            status["date"] = self.parse_datetime(
                status["created_at"], "%a %b %d %H:%M:%S %z %Y")
            status["count"] = len(files)
            yield Message.Directory, "", status

            num = 0
            for file in files:
                url = file["url"]
                if not url:
                    continue
                if url.startswith("http:"):
                    url = "https:" + url[5:]
                if "filename" not in file:
                    text.nameext_from_url(url, file)
                    if file["extension"] == "json":
                        file["extension"] = "mp4"
                if file["extension"] == "m3u8":
                    url = "ytdl:" + url
                    file["_ytdl_manifest"] = "hls"
                    file["extension"] = "mp4"
                num += 1
                file["status"] = status
                file["num"] = num
                yield Message.Url, url, file

    def _extract_status(self, status, files):
        if "mix_media_info" in status:
            for item in status["mix_media_info"]["items"]:
                type = item.get("type")
                if type == "video":
                    if self.videos:
                        files.append(self._extract_video(
                            item["data"]["media_info"]))
                elif type == "pic":
                    files.append(item["data"]["largest"].copy())
                else:
                    self.log.warning("Unknown media type '%s'", type)
            return

        if pic_ids := status.get("pic_ids"):
            pics = status["pic_infos"]
            for pic_id in pic_ids:
                pic = pics[pic_id]
                pic_type = pic.get("type")

                if pic_type == "gif" and self.gifs:
                    if self.gifs_video:
                        files.append({"url": pic["video"]})
                    else:
                        files.append(pic["largest"].copy())

                elif pic_type == "livephoto" and self.livephoto:
                    files.append(pic["largest"].copy())
                    files.append({"url": pic["video"]})

                else:
                    files.append(pic["largest"].copy())

        if "page_info" in status:
            info = status["page_info"]
            if "media_info" in info and self.videos:
                if info.get("type") != "5" or self.movies:
                    files.append(self._extract_video(info["media_info"]))
                else:
                    self.log.debug("%s: Ignoring 'movie' video", status["id"])

    def _extract_video(self, info):
        if info.get("live_status") == 1:
            self.log.debug("Skipping ongoing live stream")
            return {"url": ""}

        try:
            media = max(info["playback_list"],
                        key=lambda m: m["meta"]["quality_index"])
        except Exception:
            video = {"url": (info.get("replay_hd") or
                             info.get("stream_url_hd") or
                             info.get("stream_url") or "")}
        else:
            video = media["play_info"].copy()

        if "//wblive-out." in video["url"] and \
                not text.ext_from_url(video["url"]):
            try:
                video["url"] = self.request_location(video["url"])
            except exception.HttpError as exc:
                self.log.warning("%s: %s", exc.__class__.__name__, exc)
                video["url"] = ""

        return video

    def _status_by_id(self, status_id):
        url = (f"{self.root}/ajax/statuses/show"
               f"?id={status_id}&isGetLongText=true")
        return self.request_json(url)

    def _user(self, user):
        url = (f"{self.root}/ajax/profile/info?"
               f"{'screen_name' if self._prefix == 'n' else 'custom'}={user}")
        return self.request_json(url, interval=False)["data"]["user"]

    def _user_id(self):
        user = self.user
        if len(user) >= 10 and user.isdecimal():
            return user[-10:]
        else:
            return self._user(user)["idstr"]

    def _pagination(self, endpoint, params,
                    since_key="sinceid", subalbums=None):
        url = f"{self.root}/ajax{endpoint}"
        headers = {
            "Accept": "application/json, text/plain, */*",
            "X-Requested-With": "XMLHttpRequest",
            "X-XSRF-TOKEN": None,
            "Referer": f"{self.root}/u/{params['uid']}",
        }

        while True:
            response = self.request(url, params=params, headers=headers)
            headers["X-XSRF-TOKEN"] = response.cookies.get("XSRF-TOKEN")

            data = response.json()
            if not data.get("ok"):
                self.log.debug(response.content)
                if "since_id" not in params:  # first iteration
                    raise exception.AbortExtraction(
                        f'"{data.get("msg") or "unknown error"}"')

            try:
                data = data["data"]
                statuses = data["list"]
            except KeyError:
                return

            if subalbums is not None:
                subalbums = None
                yield data.get("album_list") or ()

            yield from statuses

            # videos, newvideo
            if cursor := data.get("next_cursor"):
                if cursor == -1:
                    return
                params["cursor"] = cursor
                continue

            # album
            if "since_id" in data:
                params[since_key] = since_id = data["since_id"]
                if not since_id:
                    return
                if "page" in params:
                    params["page"] += 1
                continue

            # home, article
            if "page" in params:
                if not statuses:
                    return
                params["page"] += 1
                continue

            # feed, last album page
            try:
                params["since_id"] = statuses[-1]["id"] - 1
            except LookupError:
                return

    def _sina_visitor_system(self, response):
        self.log.info("Sina Visitor System")

        passport_url = "https://passport.weibo.com/visitor/genvisitor"
        headers = {"Referer": response.url}
        data = {
            "cb": "gen_callback",
            "fp": '{"os":"1","browser":"Gecko109,0,0,0","fonts":"undefined",'
                  '"screenInfo":"1920*1080*24","plugins":""}',
        }

        page = Extractor.request(
            self, passport_url, method="POST", headers=headers, data=data).text
        data = util.json_loads(text.extr(page, "(", ");"))["data"]

        passport_url = "https://passport.weibo.com/visitor/visitor"
        params = {
            "a"    : "incarnate",
            "t"    : data["tid"],
            "w"    : "3" if data.get("new_tid") else "2",
            "c"    : f"{data.get('confidence') or 100:>03}",
            "gc"   : "",
            "cb"   : "cross_domain",
            "from" : "weibo",
            "_rand": random.random(),
        }
        response = Extractor.request(self, passport_url, params=params)
        _cookie_cache.update("", response.cookies)


class WeiboUserExtractor(WeiboExtractor):
    """Extractor for weibo user profiles"""
    subcategory = "user"
    pattern = USER_PATTERN + r"(?:$|#)"
    example = "https://weibo.com/USER"

    # do NOT override 'initialize()'
    # it is needed for 'self._user_id()'
    # def initialize(self):
    #     pass

    def items(self):
        base = f"{self.root}/u/{self._user_id()}?tabtype="
        return Dispatch._dispatch_extractors(self, (
            (WeiboHomeExtractor    , base + "home"),
            (WeiboFeedExtractor    , base + "feed"),
            (WeiboVideosExtractor  , base + "video"),
            (WeiboNewvideoExtractor, base + "newVideo"),
            (WeiboArticleExtractor , base + "article"),
            (WeiboAlbumExtractor   , base + "album"),
        ), ("feed",), {
            ("album", "subalbums", base + "album-only"),
        })


class WeiboHomeExtractor(WeiboExtractor):
    """Extractor for weibo 'home' listings"""
    subcategory = "home"
    pattern = USER_PATTERN + r"\?tabtype=home"
    example = "https://weibo.com/USER?tabtype=home"

    def statuses(self):
        endpoint = "/profile/myhot"
        params = {"uid": self._user_id(), "page": 1, "feature": "2"}
        return self._pagination(endpoint, params)


class WeiboFeedExtractor(WeiboExtractor):
    """Extractor for weibo user feeds"""
    subcategory = "feed"
    pattern = USER_PATTERN + r"\?tabtype=feed"
    example = "https://weibo.com/USER?tabtype=feed"

    def statuses(self):
        endpoint = "/statuses/mymblog"
        params = {"uid": self._user_id(), "feature": "0"}
        if self.logged_in:
            params["page"] = 1
        return self._pagination(endpoint, params)


class WeiboVideosExtractor(WeiboExtractor):
    """Extractor for weibo 'video' listings"""
    subcategory = "videos"
    pattern = USER_PATTERN + r"\?tabtype=video"
    example = "https://weibo.com/USER?tabtype=video"

    def statuses(self):
        endpoint = "/profile/getprofilevideolist"
        params = {"uid": self._user_id()}

        for status in self._pagination(endpoint, params):
            yield status["video_detail_vo"]


class WeiboNewvideoExtractor(WeiboExtractor):
    """Extractor for weibo 'newVideo' listings"""
    subcategory = "newvideo"
    pattern = USER_PATTERN + r"\?tabtype=newVideo"
    example = "https://weibo.com/USER?tabtype=newVideo"

    def statuses(self):
        endpoint = "/profile/getWaterFallContent"
        params = {"uid": self._user_id()}
        return self._pagination(endpoint, params)


class WeiboArticleExtractor(WeiboExtractor):
    """Extractor for weibo 'article' listings"""
    subcategory = "article"
    pattern = USER_PATTERN + r"\?tabtype=article"
    example = "https://weibo.com/USER?tabtype=article"

    def statuses(self):
        endpoint = "/statuses/mymblog"
        params = {"uid": self._user_id(), "page": 1, "feature": "10"}
        return self._pagination(endpoint, params)


class WeiboAlbumExtractor(WeiboExtractor):
    """Extractor for weibo 'album' listings"""
    subcategory = "album"
    pattern = USER_PATTERN + r"\?tabtype=album(?:[:_-]([^&#]+))?"
    example = "https://weibo.com/USER?tabtype=album"

    def items(self):
        subalbum = self.groups[2]

        if not subalbum and not self.config("subalbums", False):
            return WeiboExtractor.items(self)

        self.directory_fmt = ("{category}", "{user[screen_name]}",
                              "Album", "{subalbum[pic_title]|''}")
        self.filename_fmt = "{filename}.{extension}"
        self.archive_fmt = "{subalbum[pic_title]}_{pid}"
        return self.items_subalbum(subalbum)

    def items_subalbum(self, subalbum):
        user = self.kwdict["user"] = self._user(self.user)
        base = self.root + "/ajax/common/download?pid="

        for data, files in self.albums(user["idstr"], subalbum):
            self.kwdict["subalbum"] = data
            yield Message.Directory, "", {}
            for file in files:
                if "pid" in file:
                    file["filename"] = file["pid"]
                    file["extension"] = "jpg"
                    yield Message.Url, base + file["pid"], file
                elif "mid" in file:
                    mid = file["mid"]
                    status = self._status_by_id(mid)
                    if status.get("ok") != 1:
                        self.log.debug("Skipping status %s (%s)", mid, status)
                    else:
                        self.statuses = lambda: (status,)
                        yield from WeiboExtractor.items(self)
                        yield Message.Directory, "", {}

    def statuses(self):
        endpoint = "/profile/getImageWall"
        params = {"uid": self._user_id()}

        seen = set()
        for image in self._pagination(endpoint, params):
            mid = image["mid"]
            if mid not in seen:
                seen.add(mid)
                status = self._status_by_id(mid)
                if status.get("ok") != 1:
                    self.log.debug("Skipping status %s (%s)", mid, status)
                else:
                    yield status

    def albums(self, uid, subalbum):
        endpoint = "/profile/getImageWall"
        params = {
            "uid"      : uid,
            "sinceid"  : "0",
            "has_album": "true",
        }
        album = self._pagination(endpoint, params, subalbums=True)
        subalbums = next(album, ())

        if not subalbum or subalbum == "0":
            return (({}, album),)

        if subalbum == "all":
            results = [
                (sub, self._pagination_subalbum(uid, sub))
                for sub in subalbums
            ]
            results.append(({}, album))
            return results

        if subalbum == "only":
            return [
                (sub, self._pagination_subalbum(uid, sub))
                for sub in subalbums
            ]

        if subalbum.isdecimal():
            try:
                sub = subalbums[int(subalbum)-1]
            except Exception:
                raise exception.NotFoundError("subalbum")
        else:
            subalbum = text.unquote(subalbum)
            for sub in subalbums:
                if sub["pic_title"] == subalbum:
                    break
            else:
                raise exception.NotFoundError("subalbum")
        return ((sub, self._pagination_subalbum(uid, sub)),)

    def _pagination_subalbum(self, uid, sub):
        params = {"uid": uid, "containerid": text.unquote(sub["containerid"])}
        return self._pagination("/profile/getAlbumDetail", params, "since_id")


class WeiboStatusExtractor(WeiboExtractor):
    """Extractor for a weibo status"""
    subcategory = "status"
    pattern = BASE_PATTERN + r"/(detail|status|\d+)/(\w+)"
    example = "https://weibo.com/detail/12345"

    def statuses(self):
        status = self._status_by_id(self.user)
        if status.get("ok") != 1:
            self.log.debug(status)
            raise exception.NotFoundError("status")
        return (status,)


@cache(maxage=365*86400)
def _cookie_cache():
    return None



# -*- coding: utf-8 -*-

# Copyright 2016-2017 Leonardo Taccari
# Copyright 2017-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://www.slideshare.net/"""

from .common import GalleryExtractor


class SlidesharePresentationExtractor(GalleryExtractor):
    """Extractor for images from a presentation on slideshare.net"""
    category = "slideshare"
    subcategory = "presentation"
    directory_fmt = ("{category}", "{user}")
    filename_fmt = "{presentation}-{num:>02}.{extension}"
    archive_fmt = "{presentation}_{num}"
    pattern = (r"(?:https?://)?(?:www\.)?slideshare\.net"
               r"/(?:mobile/)?([^/?#]+)/([^/?#]+)")
    example = "https://www.slideshare.net/USER/PRESENTATION"

    def __init__(self, match):
        self.user, self.presentation = match.groups()
        url = f"https://www.slideshare.net/{self.user}/{self.presentation}"
        GalleryExtractor.__init__(self, match, url)

    def metadata(self, page):
        data = self._extract_nextdata(page)
        self.slideshow = slideshow = data["props"]["pageProps"]["slideshow"]

        return {
            "user"        : slideshow["username"],
            "presentation": self.presentation,
            "title"       : slideshow["title"].strip(),
            "description" : slideshow["description"].strip(),
            "views"       : slideshow["views"],
            "likes"       : slideshow["likes"],
            "date"        : self.parse_datetime_iso(
                slideshow["createdAt"][:19]),
        }

    def images(self, page):
        slides = self.slideshow["slides"]
        begin = (f"{slides['host']}/{slides['imageLocation']}"
                 f"/95/{slides['title']}-")
        end = "-1024.jpg"

        return [
            (begin + str(n) + end, None)
            for n in range(1, self.slideshow["totalSlides"]+1)
        ]



# -*- coding: utf-8 -*-

# Copyright 2019-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://xhamster.com/"""

from .common import Extractor, Message
from .. import text, util

BASE_PATTERN = (r"(?:https?://)?((?:[\w-]+\.)?xhamster"
                r"(?:\d?\.(?:com|one|desi)|\.porncache\.net))")


class XhamsterExtractor(Extractor):
    """Base class for xhamster extractors"""
    category = "xhamster"

    def __init__(self, match):
        self.root = "https://" + match[1]
        Extractor.__init__(self, match)


class XhamsterGalleryExtractor(XhamsterExtractor):
    """Extractor for image galleries on xhamster.com"""
    subcategory = "gallery"
    directory_fmt = ("{category}", "{user[name]}",
                     "{gallery[id]} {gallery[title]}")
    filename_fmt = "{num:>03}_{id}.{extension}"
    archive_fmt = "{id}"
    pattern = BASE_PATTERN + r"(/photos/gallery/[^/?#]+)"
    example = "https://xhamster.com/photos/gallery/12345"

    def items(self):
        data = self.metadata()
        yield Message.Directory, "", data
        for num, image in enumerate(self.images(), 1):
            url = image["imageURL"]
            image.update(data)
            text.nameext_from_url(url, image)
            image["num"] = num
            image["extension"] = "webp"
            del image["modelName"]
            yield Message.Url, url, image

    def metadata(self):
        data = self.data = self._extract_data(self.root + self.groups[1])

        gallery = data["galleryPage"]
        info = gallery["infoProps"]
        model = gallery["galleryModel"]
        author = info["authorInfoProps"]

        return {
            "user":
            {
                "id"         : text.parse_int(model["userId"]),
                "url"        : author["authorLink"],
                "name"       : author["authorName"],
                "verified"   : True if author.get("verified") else False,
                "subscribers": info["subscribeButtonProps"]["subscribers"],
            },
            "gallery":
            {
                "id"         : text.parse_int(gallery["id"]),
                "tags"       : [t["label"] for t in info["categoriesTags"]],
                "date"       : self.parse_timestamp(model["created"]),
                "views"      : text.parse_int(model["views"]),
                "likes"      : text.parse_int(model["rating"]["likes"]),
                "dislikes"   : text.parse_int(model["rating"]["dislikes"]),
                "title"      : model["title"],
                "description": model["description"],
                "thumbnail"  : model["thumbURL"],
            },
            "count": text.parse_int(gallery["photosCount"]),
        }

    def images(self):
        data = self.data
        self.data = None

        while True:
            yield from data["photosGalleryModel"]["photos"]

            pagination = data["galleryPage"]["paginationProps"]
            if pagination["currentPageNumber"] >= pagination["lastPageNumber"]:
                return
            url = (pagination["pageLinkTemplate"][:-3] +
                   str(pagination["currentPageNumber"] + 1))

            data = self._extract_data(url)

    def _extract_data(self, url):
        page = self.request(url).text
        return util.json_loads(text.extr(
            page, "window.initials=", "</script>").rstrip("\n\r;"))


class XhamsterUserExtractor(XhamsterExtractor):
    """Extractor for all galleries of an xhamster user"""
    subcategory = "user"
    pattern = BASE_PATTERN + r"/users/([^/?#]+)(?:/photos)?/?(?:$|[?#])"
    example = "https://xhamster.com/users/USER/photos"

    def items(self):
        url = f"{self.root}/users/{self.groups[1]}/photos"
        data = {"_extractor": XhamsterGalleryExtractor}

        while url:
            extr = text.extract_from(self.request(url).text)
            while True:
                url = extr(' role-pop" href="', '"')
                if not url:
                    break
                yield Message.Queue, url, data
            url = extr('data-page="next" href="', '"')



# -*- coding: utf-8 -*-

# Copyright 2023-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Run Python functions"""

from .common import PostProcessor
from .. import util


class PythonPP(PostProcessor):

    def __init__(self, job, options):
        PostProcessor.__init__(self, job)

        mode = options.get("mode")
        if mode == "eval" or not mode and options.get("expression"):
            self.function = util.compile_expression(options["expression"])
        else:
            spec = options["function"]
            module_name, _, function_name = spec.rpartition(":")
            module = util.import_file(module_name)
            self.function = getattr(module, function_name)

        if archive := self._archive_init(job, options):
            self.run = self.run_archive

        events = options.get("event")
        if events is None:
            events = ("file",)
        elif isinstance(events, str):
            events = events.split(",")
        job.register_hooks({event: self.run for event in events}, options)

        if archive:
            self._archive_register(job)

    def run(self, pathfmt):
        self.function(pathfmt.kwdict)

    def run_archive(self, pathfmt):
        kwdict = pathfmt.kwdict
        if self.archive.check(kwdict):
            return
        self.function(kwdict)
        self.archive.add(kwdict)


__postprocessor__ = PythonPP



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Generic information extractor"""

from .common import Extractor, Message
from .. import config, text
import os.path


class GenericExtractor(Extractor):
    """Extractor for images in a generic web page."""
    category = "generic"
    directory_fmt = ("{category}", "{subcategory}", "{path}")
    archive_fmt = "{imageurl}"

    # By default, the generic extractor is disabled
    # and the "g(eneric):" prefix in url is required.
    # If the extractor is enabled, make the prefix optional
    pattern = r"(?i)(?P<generic>g(?:eneric)?:)"
    if config.get(("extractor", "generic"), "enabled"):
        pattern += r"?"

    # The generic extractor pattern should match (almost) any valid url
    # Based on: https://tools.ietf.org/html/rfc3986#appendix-B
    pattern += (
        r"(?P<scheme>https?://)?"          # optional http(s) scheme
        r"(?P<domain>[-\w\.]+)"            # required domain
        r"(?P<path>/[^?#]*)?"              # optional path
        r"(?:\?(?P<query>[^#]*))?"         # optional query
        r"(?:\#(?P<fragment>.*))?"         # optional fragment
    )
    example = "generic:https://www.nongnu.org/lzip/"

    def __init__(self, match):
        self.subcategory = match['domain']
        Extractor.__init__(self, match)

        # Strip the "g(eneric):" prefix
        # and inform about "forced" or "fallback" mode
        if match['generic']:
            self.url = match[0].partition(":")[2]
        else:
            self.log.info("Falling back on generic information extractor.")
            self.url = match[0]

        # Make sure we have a scheme, or use https
        if match['scheme']:
            self.scheme = match['scheme']
        else:
            self.scheme = 'https://'
            self.url = text.ensure_http_scheme(self.url, self.scheme)

        self.path = match['path']

        # Used to resolve relative image urls
        self.root = self.scheme + match['domain']

    def items(self):
        """Get page, extract metadata & images, yield them in suitable messages

        Adapted from common.GalleryExtractor.items()

        """
        page = self.request(self.url).text
        data = self.metadata(page)
        imgs = self.images(page)

        try:
            data["count"] = len(imgs)
        except TypeError:
            pass
        images = enumerate(imgs, 1)

        yield Message.Directory, "", data

        for data["num"], (url, imgdata) in images:
            if imgdata:
                data.update(imgdata)
                if "extension" not in imgdata:
                    text.nameext_from_url(url, data)
            else:
                text.nameext_from_url(url, data)
            yield Message.Url, url, data

    def metadata(self, page):
        """Extract generic webpage metadata, return them in a dict."""
        data = {
            "title"         : text.extr(
                page, "<title>", "</title>"),
            "description"   : text.extr(
                page, '<meta name="description" content="', '"'),
            "keywords"      : text.extr(
                page, '<meta name="keywords" content="', '"'),
            "language"      : text.extr(
                page, '<meta name="language" content="', '"'),
            "name"          : text.extr(
                page, '<meta itemprop="name" content="', '"'),
            "copyright"     : text.extr(
                page, '<meta name="copyright" content="', '"'),
            "og_site"       : text.extr(
                page, '<meta property="og:site" content="', '"'),
            "og_site_name"  : text.extr(
                page, '<meta property="og:site_name" content="', '"'),
            "og_title"      : text.extr(
                page, '<meta property="og:title" content="', '"'),
            "og_description": text.extr(
                page, '<meta property="og:description" content="', '"'),

        }

        data = {k: text.unescape(v) for k, v in data.items() if v}
        data["path"] = self.path.replace("/", "")
        data["pageurl"] = self.url

        return data

    def images(self, page):
        """Extract image urls, return a list of (image url, metadata) tuples.

        The extractor aims at finding as many _likely_ image urls as possible,
        using two strategies (see below); since these often overlap, any
        duplicate urls will be removed at the end of the process.

        Note: since we are using re.findall() (see below), it's essential that
        the following patterns contain 0 or at most 1 capturing group, so that
        re.findall() return a list of urls (instead of a list of tuples of
        matching groups). All other groups used in the pattern should be
        non-capturing (?:...).

        1: Look in src/srcset attributes of img/video/source elements

        See:
        https://www.w3schools.com/tags/att_src.asp
        https://www.w3schools.com/tags/att_source_srcset.asp

        We allow both absolute and relative urls here.

        Note that srcset attributes often contain multiple space separated
        image urls; this pattern matches only the first url; remaining urls
        will be matched by the "imageurl_pattern_ext" pattern below.
        """

        imageurl_pattern_src = (
            r"(?i)"
            r"<(?:img|video|source)\s[^>]*"    # <img>, <video> or <source>
            r"src(?:set)?=[\"']?"              # src or srcset attributes
            r"(?P<URL>[^\"'\s>]+)"             # url
        )

        """
        2: Look anywhere for urls containing common image/video extensions

        The list of allowed extensions is borrowed from the directlink.py
        extractor; other could be added, see
        https://en.wikipedia.org/wiki/List_of_file_formats

        Compared to the "pattern" class variable, here we must exclude also
        other special characters (space, ", ', <, >), since we are looking for
        urls in html tags.
        """

        imageurl_pattern_ext = (
            r"(?i)"
            r"(?:[^?&#\"'>\s]+)"           # anything until dot+extension
                                           # dot + image/video extensions
            r"\.(?:jpe?g|jpe|png|gif|web[mp]|mp4|mkv|og[gmv]|opus)"
            r"(?:[^\"'<>\s]*)?"            # optional query and fragment
        )

        imageurls_src = text.re(imageurl_pattern_src).findall(page)
        imageurls_ext = text.re(imageurl_pattern_ext).findall(page)
        imageurls = imageurls_src + imageurls_ext

        # Resolve relative urls
        #
        # Image urls catched so far may be relative, so we must resolve them
        # by prepending a suitable base url.
        #
        # If the page contains a <base> element, use it as base url
        basematch = text.re(
            r"(?i)(?:<base\s.*?href=[\"']?)(?P<url>[^\"' >]+)").search(page)
        if basematch:
            self.baseurl = basematch['url'].rstrip('/')
        # Otherwise, extract the base url from self.url
        else:
            if self.url.endswith("/"):
                self.baseurl = self.url.rstrip('/')
            else:
                self.baseurl = os.path.dirname(self.url)

        # Build the list of absolute image urls
        absimageurls = []
        for u in imageurls:
            # Absolute urls are taken as-is
            if u.startswith('http'):
                absimageurls.append(u)
            # // relative urls are prefixed with current scheme
            elif u.startswith('//'):
                absimageurls.append(self.scheme + u.lstrip('/'))
            # / relative urls are prefixed with current scheme+domain
            elif u.startswith('/'):
                absimageurls.append(self.root + u)
            # other relative urls are prefixed with baseurl
            else:
                absimageurls.append(self.baseurl + '/' + u)

        # Remove duplicates
        absimageurls = dict.fromkeys(absimageurls)

        # Create the image metadata dict and add imageurl to it
        # (image filename and extension are added by items())
        images = [(u, {'imageurl': u}) for u in absimageurls]

        return images



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2025-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

import os
import sys
import unittest

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from gallery_dl import path, extractor, config  # noqa E402

KWDICT = {
    "category" : "test",
    "filename" : "file",
    "extension": "ext",
    "name"     : "test-テスト-'&>-/:~",
    "ext"      : "txt",
    "foo"      : "bar",
    "id"       : 123,
}


class TestPath(unittest.TestCase):

    def _pfmt(self, data={}, kwdict=False, extr=extractor.find("noop")):
        pathfmt = path.PathFormat(extr)

        if kwdict:
            pathfmt.set_directory({
                **(kwdict if isinstance(kwdict, dict) else KWDICT),
                **data,
            })

        return pathfmt

    def setUp(self):
        config.clear()
        path.WINDOWS = False


class TestPathObject(TestPath):

    def test_default(self):
        pfmt = self._pfmt()

        self.assertEqual(pfmt.kwdict, {})
        self.assertEqual(pfmt.delete, False)
        self.assertEqual(pfmt.filename, "")
        self.assertEqual(pfmt.extension, "")
        self.assertEqual(pfmt.directory, "")
        self.assertEqual(pfmt.realdirectory, "")
        self.assertEqual(pfmt.path, "")
        self.assertEqual(pfmt.realpath, "")
        self.assertEqual(pfmt.temppath, "")
        self.assertEqual(pfmt.basedirectory, "./gallery-dl/")
        self.assertEqual(pfmt.strip, "")

        self.assertIs(pfmt.filename_conditions, None)
        self.assertIs(pfmt.directory_conditions, None)

        self.assertTrue(callable(pfmt.extension_map))
        self.assertTrue(callable(pfmt.extension_map))
        self.assertTrue(callable(pfmt.extension_map))
        self.assertTrue(callable(pfmt.clean_segment))
        self.assertTrue(callable(pfmt.clean_path))

        self.assertTrue(callable(pfmt.filename_formatter))
        for fmt in pfmt.directory_formatters:
            self.assertTrue(callable(fmt))

    def test_str(self):
        pfmt = self._pfmt()
        self.assertEqual(str(pfmt), pfmt.realdirectory)
        self.assertEqual(str(pfmt), "")

        pfmt = self._pfmt()
        pfmt.set_directory(KWDICT)
        pfmt.set_filename(KWDICT)
        pfmt.build_path()
        self.assertEqual(str(pfmt), pfmt.realpath)
        self.assertEqual(str(pfmt), "./gallery-dl/test/file.ext")


class TestPathOptions(TestPath):

    def test_option_filename(self):
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname , "file.ext")

        config.set((), "filename", "foo.{foo}")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "foo.bar")

        config.set((), "filename", {
            "foo == 'baz'": "foo",
            "id % 2"      : "bar",
            ""            : "baz",
        })
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "bar")

    def test_option_directory(self):
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.directory    , "./gallery-dl/test/")
        self.assertEqual(pfmt.realdirectory, "./gallery-dl/test/")

        config.set((), "directory", ["{foo}", "bar"])
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.directory    , "./gallery-dl/bar/bar/")
        self.assertEqual(pfmt.realdirectory, "./gallery-dl/bar/bar/")

        config.set((), "directory", {
            "foo == 'baz'": ["a", "b", "c"],
            "id % 2"      : ["odd", "{id}"],
            ""            : ["{foo}", "bar"],
        })
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.directory    , "./gallery-dl/odd/123/")
        self.assertEqual(pfmt.realdirectory, "./gallery-dl/odd/123/")

    def test_option_basedirectory(self):
        config.set((), "base-directory", "{foo}")
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.realdirectory, "{foo}/test/")

        config.set((), "base-directory", {
            "foo == 'baz'": "bar",
            "id % 2"      : "./odd",
            ""            : "./default",
        })
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.realdirectory, "./odd/test/")

    def test_option_keywordsdefault(self):
        config.set((), "directory", ["{baz}"])
        config.set((), "base-directory", "")

        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.realdirectory, "None/")

        config.set((), "keywords-default", "ND")
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.realdirectory, "ND/")

        config.set((), "keywords-default", "")
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(pfmt.realdirectory, "")

    def test_option_extensionmap_default(self):
        kwdict = KWDICT.copy()
        pfmt = self._pfmt()
        pfmt.set_filename(kwdict)
        self.assertEqual(pfmt.extension, "ext")

        pfmt.set_extension("jpg")
        self.assertEqual(pfmt.extension, "jpg")
        self.assertEqual(kwdict["extension"], "jpg")

        pfmt.set_extension("png")
        self.assertEqual(pfmt.extension, "png")
        self.assertEqual(kwdict["extension"], "png")

        pfmt.set_extension("jpeg")
        self.assertEqual(pfmt.extension, "jpg")
        self.assertEqual(kwdict["extension"], "jpg")

        for ext, repl in path.EXTENSION_MAP.items():
            pfmt.set_extension(ext)
            self.assertEqual(pfmt.extension, repl)
            self.assertEqual(kwdict["extension"], repl)

    def test_option_extensionmap_custom(self):
        extmap = {
            "bitmap": "bmp",
            "ping"  : "png",
            "jiff"  : "gif",
        }
        config.set((), "extension-map", extmap)

        kwdict = KWDICT.copy()
        pfmt = self._pfmt()
        pfmt.set_filename(kwdict)

        pfmt.set_extension("jpg")
        self.assertEqual(pfmt.extension, "jpg")
        self.assertEqual(kwdict["extension"], "jpg")

        pfmt.set_extension("ping")
        self.assertEqual(pfmt.extension, "png")
        self.assertEqual(kwdict["extension"], "png")

        for ext, repl in extmap.items():
            pfmt.set_extension(ext)
            self.assertEqual(pfmt.extension, repl)
            self.assertEqual(kwdict["extension"], repl)

        for ext, repl in path.EXTENSION_MAP.items():
            pfmt.set_extension(ext)
            self.assertNotEqual(pfmt.extension, repl)
            self.assertNotEqual(kwdict["extension"], repl)

    def test_option_pathrestrict(self):
        config.set((), "filename", "{name}.{ext}")

        config.set((), "path-restrict", "unix")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "test-テスト-'&>-_:~.txt", "unix")

        config.set((), "path-restrict", "windows")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "test-テスト-'&_-__~.txt", "windows")

        config.set((), "path-restrict", "windows+")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "test-テスト-'&＞-⧸：~.txt", "windows+")

        config.set((), "path-restrict", "ascii")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "test____________.txt", "ascii")

        config.set((), "path-restrict", "ascii+")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "test-___-'&_-__~.txt", "ascii+")

    def test_option_pathrestrict_custom(self):
        config.set((), "filename", "{name}.{ext}")

        config.set((), "path-restrict", "ts-")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "_e___テスト_'&>_/:~._x_", "custom str")

        config.set((), "path-restrict", {
            "t": "A",
            "s": "B",
            "-": "###",
            "/": "|"
        })
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "AeBA###テスト###'&>###|:~.AxA", "custom dict")

        config.set((), "path-restrict", {
            "a-z": "x",
            "テ": "te",
            "ス": "su",
            "ト": "to",
        })
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "xxxx-tesuto-'&>-/:~.xxx", "custom dict range")

    def test_option_pathreplace(self):
        config.set((), "filename", "{name}.{ext}")

        config.set((), "path-restrict", "unix")
        config.set((), "path-replace", "&")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "test-テスト-'&>-&:~.txt", "&")

        config.set((), "path-restrict", "windows")
        config.set((), "path-replace", "***")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "test-テスト-'&***-******~.txt", "***")

    def test_option_pathremove(self):
        config.set((), "filename", "{name}.{ext}")

        config.set((), "path-remove", "-&/")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "testテスト'>_:~.txt")

        config.set((), "path-remove", "a-z0-9")
        fname = self._pfmt().build_filename(KWDICT)
        self.assertEqual(fname, "-テスト-'&>-_:~.")

    def test_option_pathstrip(self):
        config.set((), "directory", [" . {name}.{ext} . "])
        config.set((), "base-directory", "")
        config.set((), "path-restrict", "unix")

        config.set((), "path-strip", "unix")
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(
            pfmt.realdirectory, ". test-テスト-'&>-_:~.txt ./", "unix")

        config.set((), "path-strip", "windows")
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(
            pfmt.realdirectory, ". test-テスト-'&>-_:~.txt/", "windows")

        config.set((), "path-strip", "txt")
        pfmt = self._pfmt(kwdict=True)
        self.assertEqual(
            pfmt.realdirectory, ". test-テスト-'&>-_:~.txt ./", "custom")


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

# Copyright 2017-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Utility functions and classes"""

import os
import sys
import json
import time
import random
import getpass
import hashlib
import binascii
import functools
import itertools
import subprocess
import collections
import urllib.parse
from http.cookiejar import Cookie
from email.utils import mktime_tz, parsedate_tz
from . import text, dt, version, exception


def bencode(num, alphabet="0123456789"):
    """Encode an integer into a base-N encoded string"""
    data = ""
    base = len(alphabet)
    while num:
        num, remainder = divmod(num, base)
        data = alphabet[remainder] + data
    return data


def bdecode(data, alphabet="0123456789"):
    """Decode a base-N encoded string ( N = len(alphabet) )"""
    num = 0
    base = len(alphabet)
    for c in data:
        num = num * base + alphabet.find(c)
    return num


def decrypt_xor(encrypted, key, base64=True, fromhex=False):
    if base64:
        encrypted = binascii.a2b_base64(encrypted)
    if fromhex:
        encrypted = bytes.fromhex(encrypted.decode())

    div = len(key)
    return bytes([
        encrypted[i] ^ key[i % div]
        for i in range(len(encrypted))
    ]).decode()


def advance(iterable, num):
    """"Advance 'iterable' by 'num' steps"""
    iterator = iter(iterable)
    next(itertools.islice(iterator, num, num), None)
    return iterator


def repeat(times):
    """Return an iterator that returns None"""
    if times < 0:
        return itertools.repeat(None)
    return itertools.repeat(None, times)


def unique(iterable):
    """Yield unique elements from 'iterable' while preserving order"""
    seen = set()
    add = seen.add
    for element in iterable:
        if element not in seen:
            add(element)
            yield element


def unique_sequence(iterable):
    """Yield sequentially unique elements from 'iterable'"""
    last = None
    for element in iterable:
        if element != last:
            last = element
            yield element


def contains(values, elements, separator=" "):
    """Returns True if at least one of 'elements' is contained in 'values'"""
    if isinstance(values, str) and (separator or separator is None):
        values = values.split(separator)

    if not isinstance(elements, (tuple, list)):
        return elements in values

    for e in elements:
        if e in values:
            return True
    return False


def raises(cls):
    """Returns a function that raises 'cls' as exception"""
    def wrap(*args):
        raise cls(*args)
    return wrap


def identity(x, _=None):
    """Returns its argument"""
    return x


def true(_, __=None):
    """Always returns True"""
    return True


def false(_, __=None):
    """Always returns False"""
    return False


def noop(_=None):
    """Does nothing"""


def md5(s):
    """Generate MD5 hexdigest of 's'"""
    if not s:
        s = b""
    elif isinstance(s, str):
        s = s.encode()
    return hashlib.md5(s).hexdigest()


def sha1(s):
    """Generate SHA1 hexdigest of 's'"""
    if not s:
        s = b""
    elif isinstance(s, str):
        s = s.encode()
    return hashlib.sha1(s).hexdigest()


def generate_token(size=16):
    """Generate a random token with hexadecimal digits"""
    return random.getrandbits(size * 8).to_bytes(size, "big").hex()


def format_value(value, suffixes="kMGTPEZY"):
    value = str(value)
    value_len = len(value)
    index = value_len - 4
    if index >= 0:
        offset = (value_len - 1) % 3 + 1
        return (f"{value[:offset]}.{value[offset:offset+2]}"
                f"{suffixes[index // 3]}")
    return value


def combine_dict(a, b):
    """Recursively combine the contents of 'b' into 'a'"""
    for key, value in b.items():
        if key in a and isinstance(value, dict) and isinstance(a[key], dict):
            combine_dict(a[key], value)
        else:
            a[key] = value
    return a


def transform_dict(a, func):
    """Recursively apply 'func' to all values in 'a'"""
    for key, value in a.items():
        if isinstance(value, dict):
            transform_dict(value, func)
        else:
            a[key] = func(value)


def filter_dict(a):
    """Return a copy of 'a' without "private" entries"""
    return {k: v for k, v in a.items() if k[0] != "_"}


def delete_items(obj, keys):
    """Remove all 'keys' from 'obj'"""
    for key in keys:
        if key in obj:
            del obj[key]


def enumerate_reversed(iterable, start=0, length=None):
    """Enumerate 'iterable' and return its elements in reverse order"""
    if length is None:
        length = len(iterable)

    try:
        iterable = zip(range(start-1+length, start-1, -1), reversed(iterable))
    except TypeError:
        iterable = list(zip(range(start, start+length), iterable))
        iterable.reverse()

    return iterable


def number_to_string(value, numbers=(int, float)):
    """Convert numbers (int, float) to string; Return everything else as is."""
    return str(value) if value.__class__ in numbers else value


def to_string(value):
    """str() with "better" defaults"""
    if not value:
        return ""
    if value.__class__ is list:
        try:
            return ", ".join(value)
        except Exception:
            return ", ".join(map(str, value))
    return str(value)


def json_default(obj):
    if isinstance(obj, CustomNone):
        return None
    return str(obj)


json_loads = json._default_decoder.decode
json_dumps = json.JSONEncoder(
    check_circular=False,
    separators=(",", ":"),
    default=json_default,
).encode


def dump_json(obj, fp=sys.stdout, ensure_ascii=True, indent=4):
    """Serialize 'obj' as JSON and write it to 'fp'"""
    json.dump(
        obj, fp,
        ensure_ascii=ensure_ascii,
        indent=indent,
        default=json_default,
        sort_keys=True,
    )
    fp.write("\n")


def dump_response(response, fp, headers=False, content=True, hide_auth=True):
    """Write the contents of 'response' into a file-like object"""

    if headers:
        request = response.request
        req_headers = request.headers.copy()
        res_headers = response.headers.copy()

        if hide_auth:
            if authorization := req_headers.get("Authorization"):
                atype, sep, _ = str(authorization).partition(" ")
                req_headers["Authorization"] = f"{atype} ***" if sep else "***"

            if cookie := req_headers.get("Cookie"):
                req_headers["Cookie"] = ";".join(
                    c.partition("=")[0] + "=***"
                    for c in cookie.split(";")
                )

            if set_cookie := res_headers.get("Set-Cookie"):
                res_headers["Set-Cookie"] = re(r"(^|, )([^ =]+)=[^,;]*").sub(
                    r"\1\2=***", set_cookie)

        request_headers = "\n".join(
            f"{name}: {value}"
            for name, value in req_headers.items()
        )
        response_headers = "\n".join(
            f"{name}: {value}"
            for name, value in res_headers.items()
        )

        output = f"""\
{request.method} {request.url}
Status: {response.status_code} {response.reason}

Request Headers
---------------
{request_headers}
"""
        if request.body:
            output = f"""{output}
Request Body
------------
{request.body}
"""
        output = f"""{output}
Response Headers
----------------
{response_headers}
"""
        fp.write(output.encode())

    if content:
        if headers:
            fp.write(b"\nContent\n-------\n")
        fp.write(response.content)


def extract_headers(response):
    headers = response.headers
    data = dict(headers)

    if hcd := headers.get("content-disposition"):
        if name := text.extr(hcd, 'filename="', '"'):
            text.nameext_from_url(name, data)

    if hlm := headers.get("last-modified"):
        data["date"] = dt.datetime(*parsedate_tz(hlm)[:6])

    return data


def detect_challenge(response):
    server = response.headers.get("server")
    if not server:
        return

    elif server.startswith("cloudflare"):
        if response.status_code not in (403, 503):
            return

        mitigated = response.headers.get("cf-mitigated")
        if mitigated and mitigated.lower() == "challenge":
            return "Cloudflare challenge"

        content = response.content
        if b"_cf_chl_opt" in content or b"jschl-answer" in content:
            return "Cloudflare challenge"
        elif b'name="captcha-bypass"' in content:
            return "Cloudflare CAPTCHA"

    elif server.startswith("ddos-guard"):
        if response.status_code == 403 and \
                b"/ddos-guard/js-challenge/" in response.content:
            return "DDoS-Guard challenge"


@functools.lru_cache(maxsize=None)
def git_head():
    try:
        out, err = Popen(
            ("git", "rev-parse", "--short", "HEAD"),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=os.path.dirname(os.path.abspath(__file__)),
        ).communicate()
        if out and not err:
            return out.decode().rstrip()
    except (OSError, subprocess.SubprocessError):
        pass
    return None


def expand_path(path):
    """Expand environment variables and tildes (~)"""
    if not path:
        return path
    if not isinstance(path, str):
        path = os.path.join(*path)
    return os.path.expandvars(os.path.expanduser(path))


def remove_file(path):
    try:
        os.unlink(path)
    except OSError:
        pass


def remove_directory(path):
    try:
        os.rmdir(path)
    except OSError:
        pass


def set_mtime(path, mtime):
    try:
        if isinstance(mtime, str):
            mtime = mktime_tz(parsedate_tz(mtime))
        os.utime(path, (time.time(), mtime))
    except Exception:
        pass


def cookiestxt_load(fp):
    """Parse a Netscape cookies.txt file and add return its Cookies"""
    cookies = []

    for line in fp:

        line = line.lstrip(" ")
        # strip '#HttpOnly_'
        if line.startswith("#HttpOnly_"):
            line = line[10:]
        # ignore empty lines and comments
        if not line or line[0] in ("#", "$", "\n"):
            continue
        # strip trailing '\n'
        if line[-1] == "\n":
            line = line[:-1]

        domain, domain_specified, path, secure, expires, name, value = \
            line.split("\t")

        if not name:
            name = value
            value = None

        cookies.append(Cookie(
            0, name, value,
            None, False,
            domain,
            domain_specified == "TRUE",
            domain[0] == "." if domain else False,
            path, False,
            secure == "TRUE",
            None if expires == "0" or not expires else expires,
            False, None, None, {},
        ))

    return cookies


def cookiestxt_store(fp, cookies):
    """Write 'cookies' in Netscape cookies.txt format to 'fp'"""
    fp.write("# Netscape HTTP Cookie File\n\n")

    for cookie in cookies:
        if not cookie.domain:
            continue

        if cookie.value is None:
            name = ""
            value = cookie.name
        else:
            name = cookie.name
            value = cookie.value

        domain = cookie.domain
        fp.write(
            f"{domain}\t"
            f"{'TRUE' if domain and domain[0] == '.' else 'FALSE'}\t"
            f"{cookie.path}\t"
            f"{'TRUE' if cookie.secure else 'FALSE'}\t"
            f"{'0' if cookie.expires is None else str(cookie.expires)}\t"
            f"{name}\t"
            f"{value}\n"
        )


def code_to_language(code, default=None):
    """Map an ISO 639-1 language code to its actual name"""
    return CODES.get((code or "").lower(), default)


def language_to_code(lang, default=None):
    """Map a language name to its ISO 639-1 code"""
    if lang is None:
        return default
    lang = lang.capitalize()
    for code, language in CODES.items():
        if language == lang:
            return code
    return default


CODES = {
    "ar": "Arabic",
    "bg": "Bulgarian",
    "bn": "Bengali",
    "ca": "Catalan",
    "cs": "Czech",
    "da": "Danish",
    "de": "German",
    "el": "Greek",
    "en": "English",
    "es": "Spanish",
    "fa": "Persian",
    "fi": "Finnish",
    "fr": "French",
    "he": "Hebrew",
    "hi": "Hindi",
    "hu": "Hungarian",
    "id": "Indonesian",
    "it": "Italian",
    "ja": "Japanese",
    "ko": "Korean",
    "ms": "Malay",
    "nl": "Dutch",
    "no": "Norwegian",
    "pl": "Polish",
    "pt": "Portuguese",
    "ro": "Romanian",
    "ru": "Russian",
    "sk": "Slovak",
    "sl": "Slovenian",
    "sr": "Serbian",
    "sv": "Swedish",
    "th": "Thai",
    "tr": "Turkish",
    "uk": "Ukrainian",
    "vi": "Vietnamese",
    "zh": "Chinese",
}


def HTTPBasicAuth(username, password):
    authorization = b"Basic " + binascii.b2a_base64(
        f"{username}:{password}".encode("latin1"), newline=False)
    del username, password

    def _apply(request):
        request.headers["Authorization"] = authorization
        return request
    return _apply


class ModuleProxy():
    __slots__ = ()

    def __getitem__(self, key, modules=sys.modules):
        try:
            return modules[key]
        except KeyError:
            pass
        try:
            __import__(key)
        except ImportError:
            modules[key] = NONE
            return NONE
        return modules[key]

    __getattr__ = __getitem__


class LazyPrompt():
    __slots__ = ()

    def __str__(self):
        return getpass.getpass()


class NullContext():
    __slots__ = ()

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class NullResponse():
    __slots__ = ("url", "reason")

    ok = is_redirect = is_permanent_redirect = False
    cookies = headers = history = links = {}
    encoding = apparent_encoding = "utf-8"
    content = b""
    text = ""
    status_code = 900
    close = noop

    def __init__(self, url, reason=""):
        self.url = url
        self.reason = str(reason)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __str__(self):
        return "900 " + self.reason

    def json(self):
        return {}


class CustomNone():
    """None-style type that supports more operations than regular None"""
    __slots__ = ()

    __getattribute__ = identity
    __getitem__ = identity
    __iter__ = identity

    def __call__(self, *args, **kwargs):
        return self

    def __next__(self):
        raise StopIteration

    def __eq__(self, other):
        return other is self or other is None

    def __ne__(self, other):
        return other is not self and other is not None

    __lt__ = true
    __le__ = true
    __gt__ = false
    __ge__ = false
    __bool__ = false

    __add__ = identity
    __sub__ = identity
    __mul__ = identity
    __matmul__ = identity
    __truediv__ = identity
    __floordiv__ = identity
    __mod__ = identity

    __radd__ = identity
    __rsub__ = identity
    __rmul__ = identity
    __rmatmul__ = identity
    __rtruediv__ = identity
    __rfloordiv__ = identity
    __rmod__ = identity

    __lshift__ = identity
    __rshift__ = identity
    __and__ = identity
    __xor__ = identity
    __or__ = identity

    __rlshift__ = identity
    __rrshift__ = identity
    __rand__ = identity
    __rxor__ = identity
    __ror__ = identity

    __neg__ = identity
    __pos__ = identity
    __abs__ = identity
    __invert__ = identity

    def __len__(self):
        return 0

    __int__ = __len__
    __hash__ = __len__
    __index__ = __len__

    def __format__(self, _):
        return "None"

    def __str__(self):
        return "None"

    __repr__ = __str__


class Flags():

    def __init__(self):
        self.FILE = self.POST = self.CHILD = self.DOWNLOAD = None

    def process(self, flag):
        value = self.__dict__[flag]
        self.__dict__[flag] = None

        if value == "abort":
            raise exception.AbortExtraction()
        if value == "terminate":
            raise exception.TerminateExtraction()
        if value == "restart":
            raise exception.RestartExtraction()
        raise exception.StopExtraction()


# v137.0 release of Firefox on 2025-04-01 has ordinal 739342
# 735506 == 739342 - 137 * 28
# v135.0 release of Chrome  on 2025-04-01 has ordinal 739342
# 735562 == 739342 - 135 * 28
#  _ord_today = dt.date.today().toordinal()
#  _ff_ver = (_ord_today - 735506) // 28
#  _ch_ver = (_ord_today - 735562) // 28

_ord_today = dt.date.today().toordinal()
_ff_ver = (_ord_today - 735_513) // 28  # 147 on 2026-01-13
_ch_ver = (_ord_today - 735_599) // 28  # 143 on 2025-12-18

re = text.re
re_compile = text.re_compile

NONE = CustomNone()
FLAGS = Flags()
WINDOWS = (os.name == "nt")
SENTINEL = object()
EXECUTABLE = getattr(sys, "frozen", False)
SPECIAL_EXTRACTORS = {"oauth", "recursive", "generic"}

EXTS_IMAGE = {"jpg", "jpeg", "png", "gif", "bmp", "svg", "psd", "ico",
              "webp", "avif", "heic", "heif"}
EXTS_VIDEO = {"mp4", "m4v", "mov", "webm", "mkv", "ogv", "flv", "avi", "wmv"}
EXTS_ARCHIVE = {"zip", "rar", "7z", "tar", "gz", "bz2", "lzma", "xz"}

USERAGENT_GALLERYDL = "gallery-dl/" + version.__version__
USERAGENT_FIREFOX = (f"Mozilla/5.0 (Windows NT 10.0; Win64; x64; "
                     f"rv:{_ff_ver}.0) Gecko/20100101 Firefox/{_ff_ver}.0")
USERAGENT_CHROME = ("Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                    "AppleWebKit/537.36 (KHTML, like Gecko) "
                    f"Chrome/{_ch_ver}.0.0.0 Safari/537.36")

GLOBALS = {
    "contains" : contains,
    "parse_int": text.parse_int,
    "urlsplit" : urllib.parse.urlsplit,
    "datetime" : dt.datetime,
    "timedelta": dt.timedelta,
    "abort"    : raises(exception.StopExtraction),
    "error"    : raises(exception.AbortExtraction),
    "terminate": raises(exception.TerminateExtraction),
    "restart"  : raises(exception.RestartExtraction),
    "hash_sha1": sha1,
    "hash_md5" : md5,
    "std"      : ModuleProxy(),
    "re"       : text.re_module,
    "exts_image"  : EXTS_IMAGE,
    "exts_video"  : EXTS_VIDEO,
    "exts_archive": EXTS_ARCHIVE,
}


if EXECUTABLE and hasattr(sys, "_MEIPASS"):
    # https://github.com/pyinstaller/pyinstaller/blob/develop/doc
    # /runtime-information.rst#ld_library_path--libpath-considerations
    _popen_env = os.environ.copy()

    orig = _popen_env.get("LD_LIBRARY_PATH_ORIG")
    if orig is None:
        _popen_env.pop("LD_LIBRARY_PATH", None)
    else:
        _popen_env["LD_LIBRARY_PATH"] = orig

    orig = _popen_env.get("DYLD_LIBRARY_PATH_ORIG")
    if orig is None:
        _popen_env.pop("DYLD_LIBRARY_PATH", None)
    else:
        _popen_env["DYLD_LIBRARY_PATH"] = orig

    del orig

    def Popen(args, **kwargs):
        kwargs["env"] = _popen_env
        return subprocess.Popen(args, **kwargs)
else:
    Popen = subprocess.Popen


def compile_expression_raw(expr, name="<expr>", globals=None):
    code_object = compile(expr, name, "eval")
    return functools.partial(eval, code_object, globals or GLOBALS)


def compile_expression_defaultdict(expr, name="<expr>", globals=None):
    global GLOBALS_DEFAULT

    if isinstance(__builtins__, dict):
        # cpython
        GLOBALS_DEFAULT = collections.defaultdict(lambda n=NONE: n, GLOBALS)
    else:
        # pypy3 - insert __builtins__ symbols into globals dict
        GLOBALS_DEFAULT = collections.defaultdict(
            lambda n=NONE: n, __builtins__.__dict__)
        GLOBALS_DEFAULT.update(GLOBALS)

    global compile_expression_defaultdict
    compile_expression_defaultdict = compile_expression_defaultdict_impl
    return compile_expression_defaultdict_impl(expr, name, globals)


def compile_expression_defaultdict_impl(expr, name="<expr>", globals=None):
    code_object = compile(expr, name, "eval")
    return functools.partial(eval, code_object, globals or GLOBALS_DEFAULT)


def compile_expression_tryexcept(expr, name="<expr>", globals=None):
    code_object = compile(expr, name, "eval")
    if globals is None:
        globals = GLOBALS

    def _eval(locals=None):
        try:
            return eval(code_object, globals, locals)
        except exception.GalleryDLException:
            raise
        except Exception:
            return NONE

    return _eval


compile_expression = compile_expression_tryexcept


def compile_filter(expr, name="<filter>", globals=None):
    if not isinstance(expr, str):
        expr = f"({') and ('.join(expr)})"
    return compile_expression(expr, name, globals)


def import_file(path):
    """Import a Python module from a filesystem path"""
    path, name = os.path.split(path)

    name, sep, ext = name.rpartition(".")
    if not sep:
        name = ext

    if path:
        path = expand_path(path)
        sys.path.insert(0, path)
        try:
            return __import__(name)
        finally:
            del sys.path[0]
    else:
        return __import__(name.replace("-", "_"))


def build_selection_func(value, min=0.0, conv=float):
    if not value:
        if min:
            return lambda: min
        return None

    if isinstance(value, str):
        lower, _, upper = value.partition("-")
    else:
        try:
            lower, upper = value
        except TypeError:
            lower, upper = value, None
    lower = conv(lower)

    if upper:
        upper = conv(upper)
        return functools.partial(
            random.uniform if lower.__class__ is float else random.randint,
            lower if lower > min else min,
            upper if upper > min else min,
        )
    else:
        if lower < min:
            lower = min
        return lambda: lower


build_duration_func = build_selection_func


def build_extractor_filter(categories, negate=True, special=None):
    """Build a function that takes an Extractor class as argument
    and returns True if that class is allowed by 'categories'
    """
    if isinstance(categories, str):
        categories = categories.split(",")

    catset = set()  # set of categories / basecategories
    subset = set()  # set of subcategories
    catsub = []     # list of category-subcategory pairs

    for item in categories:
        category, _, subcategory = item.partition(":")
        if category and category != "*":
            if subcategory and subcategory != "*":
                catsub.append((category, subcategory))
            else:
                catset.add(category)
        elif subcategory and subcategory != "*":
            subset.add(subcategory)

    if special:
        catset |= special
    elif not catset and not subset and not catsub:
        return true if negate else false

    tests = []

    if negate:
        if catset:
            tests.append(lambda extr:
                         extr.category not in catset and
                         extr.basecategory not in catset)
        if subset:
            tests.append(lambda extr: extr.subcategory not in subset)
    else:
        if catset:
            tests.append(lambda extr:
                         extr.category in catset or
                         extr.basecategory in catset)
        if subset:
            tests.append(lambda extr: extr.subcategory in subset)

    if catsub:
        def test(extr):
            for category, subcategory in catsub:
                if subcategory == extr.subcategory and (
                        category == extr.category or
                        category == extr.basecategory):
                    return not negate
            return negate
        tests.append(test)

    if len(tests) == 1:
        return tests[0]
    if negate:
        return lambda extr: all(t(extr) for t in tests)
    else:
        return lambda extr: any(t(extr) for t in tests)


def build_proxy_map(proxies, log=None):
    """Generate a proxy map"""
    if not proxies:
        return None

    if isinstance(proxies, str):
        if "://" not in proxies:
            proxies = "http://" + proxies.lstrip("/")
        proxies = {"http": proxies, "https": proxies}
    elif isinstance(proxies, dict):
        for scheme, proxy in proxies.items():
            if "://" not in proxy:
                proxies[scheme] = "http://" + proxy.lstrip("/")
    else:
        proxies = None

    if log is not None:
        if proxies is None:
            log.warning("Invalid proxy specifier: %r", proxies)
        else:
            log.debug("Proxy Map: %s", proxies)

    return proxies


def predicate_build(predicates):
    if not predicates:
        return true

    if len(predicates) == 1:
        return predicates[0]

    def chain(url, kwdict):
        for pred in predicates:
            if not pred(url, kwdict):
                return False
        return True
    return chain


def predicate_unique():
    """Predicate; True if given URL has not been encountered before"""
    def _pred(url, _):
        if url.startswith("text:"):
            return True
        if url not in urls:
            urls.add(url)
            return True
        return False
    urls = set()
    return _pred


def predicate_filter(expr, target="image"):
    """Predicate; True if evaluating the given expression returns True"""
    def _pred(_, kwdict):
        try:
            return expr(kwdict)
        except exception.GalleryDLException:
            raise
        except Exception as exc:
            raise exception.FilterError(exc)
    expr = compile_filter(expr, f"<{target} filter>")
    return _pred


def predicate_range(ranges, skip=None):
    """Predicate; True if the current index is in the given range(s)"""
    if ranges := predicate_range_parse(ranges):
        # technically wrong for 'step > 2', but good enough for now
        # and evaluating min/max for a large range is slow
        upper = max(r.stop for r in ranges) - 1
        lower = min(r.start for r in ranges)
        index = 0 if skip is None or lower <= 1 else skip(lower)
        del lower
    else:
        index = upper = 0

    def _pred(_url, _kwdict):
        nonlocal index

        if index >= upper:
            raise exception.StopExtraction()
        index += 1

        for range in ranges:
            if index in range:
                return True
        return False
    return _pred


def predicate_range_parse(rangespec):
    """Parse an integer range string and return the resulting ranges

    Examples:
        _parse("-2,4,6-8,10-")      -> [(1,3), (4,5), (6,9), (10,INTMAX)]
        _parse(" - 3 , 4-  4, 2-6") -> [(1,4), (4,5), (2,7)]
        _parse("1:2,4:8:2")         -> [(1,1), (4,7,2)]
    """
    ranges = []

    if isinstance(rangespec, str):
        rangespec = rangespec.split(",")
    elif isinstance(rangespec, int):
        rangespec = (str(rangespec),)

    for group in rangespec:
        if not group:
            continue

        elif ":" in group:
            start, _, stop = group.partition(":")
            stop, _, step = stop.partition(":")
            ranges.append(range(
                int(start) if start.strip() else 1,
                int(stop) if stop.strip() else sys.maxsize,
                int(step) if step.strip() else 1,
            ))

        elif "-" in group:
            start, _, stop = group.partition("-")
            ranges.append(range(
                int(start) if start.strip() else 1,
                int(stop) + 1 if stop.strip() else sys.maxsize,
            ))

        else:
            start = int(group)
            ranges.append(range(start, start+1))

    return ranges



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://hatenablog.com"""

from .common import Extractor, Message
from .. import text


BASE_PATTERN = (
    r"(?:hatenablog:https?://([^/?#]+)|(?:https?://)?"
    r"([\w-]+\.(?:hatenablog\.(?:com|jp)"
    r"|hatenadiary\.com|hateblo\.jp)))"
)
QUERY_RE = r"(?:\?([^#]*))?(?:#.*)?$"


class HatenablogExtractor(Extractor):
    """Base class for HatenaBlog extractors"""
    category = "hatenablog"
    directory_fmt = ("{category}", "{domain}")
    filename_fmt = "{category}_{domain}_{entry}_{num:>02}.{extension}"
    archive_fmt = "{filename}"

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.domain = match[1] or match[2]

    def _init(self):
        self._find_img = text.re(r'<img +([^>]+)').finditer

    def _handle_article(self, article: str):
        extr = text.extract_from(article)
        date = self.parse_datetime_iso(extr('<time datetime="', '"'))
        entry_link = text.unescape(extr('<a href="', '"'))
        entry = entry_link.partition("/entry/")[2]
        title = text.unescape(extr('>', '<'))
        content = extr(
            '<div class="entry-content hatenablog-entry">', '</div>')

        images = []
        for match in self._find_img(content):
            attributes = match[1]
            if 'class="hatena-fotolife"' not in attributes:
                continue
            image = text.unescape(text.extr(attributes, 'src="', '"'))
            images.append(image)

        data = {
            "domain": self.domain,
            "date": date,
            "entry": entry,
            "title": title,
            "count": len(images),
        }
        yield Message.Directory, "", data
        for data["num"], url in enumerate(images, 1):
            yield Message.Url, url, text.nameext_from_url(url, data)


class HatenablogEntriesExtractor(HatenablogExtractor):
    """Base class for a list of entries"""
    allowed_parameters = ()

    def __init__(self, match):
        HatenablogExtractor.__init__(self, match)
        self.path = match[3]
        self.query = {key: value for key, value in text.parse_query(
            match[4]).items() if self._acceptable_query(key)}

    def _init(self):
        HatenablogExtractor._init(self)
        self._find_pager_url = text.re(
            r' class="pager-next">\s*<a href="([^"]+)').search

    def items(self):
        url = "https://" + self.domain + self.path
        query = self.query

        while url:
            page = self.request(url, params=query).text

            extr = text.extract_from(page)
            attributes = extr('<body ', '>')
            if "page-archive" in attributes:
                yield from self._handle_partial_articles(extr)
            else:
                yield from self._handle_full_articles(extr)

            match = self._find_pager_url(page)
            url = text.unescape(match[1]) if match else None
            query = None

    def _handle_partial_articles(self, extr):
        while True:
            section = extr('<section class="archive-entry', '</section>')
            if not section:
                break

            url = "hatenablog:" + text.unescape(text.extr(
                section, '<a class="entry-title-link" href="', '"'))
            data = {"_extractor": HatenablogEntryExtractor}
            yield Message.Queue, url, data

    def _handle_full_articles(self, extr):
        while True:
            attributes = extr('<article ', '>')
            if not attributes:
                break
            if "no-entry" in attributes:
                continue

            article = extr('', '</article>')
            yield from self._handle_article(article)

    def _acceptable_query(self, key):
        return key == "page" or key in self.allowed_parameters


class HatenablogEntryExtractor(HatenablogExtractor):
    """Extractor for a single entry URL"""
    subcategory = "entry"
    pattern = BASE_PATTERN + r"/entry/([^?#]+)" + QUERY_RE
    example = "https://BLOG.hatenablog.com/entry/PATH"

    def __init__(self, match):
        HatenablogExtractor.__init__(self, match)
        self.path = match[3]

    def items(self):
        url = "https://" + self.domain + "/entry/" + self.path
        page = self.request(url).text

        extr = text.extract_from(page)
        while True:
            attributes = extr('<article ', '>')
            if "no-entry" in attributes:
                continue
            article = extr('', '</article>')
            return self._handle_article(article)


class HatenablogHomeExtractor(HatenablogEntriesExtractor):
    """Extractor for a blog's home page"""
    subcategory = "home"
    pattern = BASE_PATTERN + r"(/?)" + QUERY_RE
    example = "https://BLOG.hatenablog.com"


class HatenablogArchiveExtractor(HatenablogEntriesExtractor):
    """Extractor for a blog's archive page"""
    subcategory = "archive"
    pattern = (BASE_PATTERN + r"(/archive(?:/\d+(?:/\d+(?:/\d+)?)?"
               r"|/category/[^?#]+)?)" + QUERY_RE)
    example = "https://BLOG.hatenablog.com/archive/2024"


class HatenablogSearchExtractor(HatenablogEntriesExtractor):
    """Extractor for a blog's search results"""
    subcategory = "search"
    pattern = BASE_PATTERN + r"(/search)" + QUERY_RE
    example = "https://BLOG.hatenablog.com/search?q=QUERY"
    allowed_parameters = ("q",)



# -*- coding: utf-8 -*-

# Copyright 2016-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://readcomiconline.li/"""

from .common import Extractor, ChapterExtractor, MangaExtractor, Message
from .. import text, exception
import binascii

BASE_PATTERN = r"(?i)(?:https?://)?(?:www\.)?readcomiconline\.(?:li|to)"


class ReadcomiconlineBase():
    """Base class for readcomiconline extractors"""
    category = "readcomiconline"
    directory_fmt = ("{category}", "{comic}", "{issue:>03}")
    filename_fmt = "{comic}_{issue:>03}_{page:>03}.{extension}"
    archive_fmt = "{issue_id}_{page}"
    root = "https://readcomiconline.li"
    request_interval = (3.0, 6.0)

    def request(self, url, **kwargs):
        """Detect and handle redirects to CAPTCHA pages"""
        while True:
            response = Extractor.request(self, url, **kwargs)
            if not response.history or "/AreYouHuman" not in response.url:
                return response
            if self.config("captcha", "stop") == "wait":
                self.log.warning(
                    "Redirect to \n%s\nVisit this URL in your browser, solve "
                    "the CAPTCHA, and press ENTER to continue", response.url)
                self.input()
            else:
                raise exception.AbortExtraction(
                    f"Redirect to \n{response.url}\nVisit this URL in your "
                    f"browser and solve the CAPTCHA to continue")


class ReadcomiconlineIssueExtractor(ReadcomiconlineBase, ChapterExtractor):
    """Extractor for comic-issues from readcomiconline.li"""
    subcategory = "issue"
    pattern = BASE_PATTERN + r"(/Comic/[^/?#]+/[^/?#]+\?)([^#]+)"
    example = "https://readcomiconline.li/Comic/TITLE/Issue-123?id=12345"

    def _init(self):
        params = text.parse_query(self.groups[1])
        quality = self.config("quality")

        if quality is None or quality == "auto":
            if "quality" not in params:
                params["quality"] = "hq"
        else:
            params["quality"] = str(quality)
        params["readType"] = "0"  # force "One page" Reading mode (#7890)

        self.page_url += "&".join(f"{k}={v}" for k, v in params.items())
        self.issue_id = params.get("id")

    def metadata(self, page):
        comic, pos = text.extract(page, "   - Read\r\n    ", "\r\n")
        iinfo, pos = text.extract(page, "    ", "\r\n", pos)
        match = text.re(r"(?:Issue )?#(\d+)|(.+)").match(iinfo)
        return {
            "comic": comic,
            "issue": match[1] or match[2],
            "issue_id": text.parse_int(self.issue_id),
            "lang": "en",
            "language": "English",
        }

    def images(self, page):
        results = []
        referer = {"_http_headers": {"Referer": self.page_url}}
        root, pos = text.extract(page, "return baeu(l, '", "'")
        _   , pos = text.extract(page, "var pth = '", "", pos)
        var , pos = text.extract(page, "var ", "= '", pos)

        replacements = text.re(
            r"l = l\.replace\(/([^/]+)/g, [\"']([^\"']*)").findall(page)

        for path in page.split(var)[2:]:
            path = text.extr(path, "= '", "'")

            for needle, repl in replacements:
                path = path.replace(needle, repl)

            results.append((baeu(path, root), referer))

        return results


class ReadcomiconlineComicExtractor(ReadcomiconlineBase, MangaExtractor):
    """Extractor for comics from readcomiconline.li"""
    chapterclass = ReadcomiconlineIssueExtractor
    subcategory = "comic"
    pattern = BASE_PATTERN + r"(/Comic/[^/?#]+/?)$"
    example = "https://readcomiconline.li/Comic/TITLE"

    def chapters(self, page):
        results = []
        comic, pos = text.extract(page, ' class="barTitle">', '<')
        page , pos = text.extract(page, ' class="listing">', '</table>', pos)

        comic = comic.rpartition("information")[0].strip()
        needle = f' title="Read {comic} '
        comic = text.unescape(comic)

        for item in text.extract_iter(page, ' href="', ' comic online '):
            url, _, issue = item.partition(needle)
            url = url.rpartition('"')[0]
            if issue.startswith('Issue #'):
                issue = issue[7:]
            results.append((self.root + url, {
                "comic": comic, "issue": issue,
                "issue_id": text.parse_int(url.rpartition("=")[2]),
                "lang": "en", "language": "English",
            }))
        return results


class ReadcomiconlineTagExtractor(ReadcomiconlineBase, Extractor):
    """Extractor for comics from readcomiconline.li lists"""
    subcategory = "tag"
    pattern = (BASE_PATTERN +
               r"(/(Artist|Genre|Publisher|Writer|)/([^/?#]+)(?:/[^/?#]+)?)")
    example = "https://readcomiconline.li/Artist/NAME"

    def __init__(self, match):
        self.subcategory = match[2].lower()
        Extractor.__init__(self, match)

    def items(self):
        path, _, name = self.groups
        self.kwdict["search_tags"] = text.unquote(name)

        self.cookies.set("list-view", "list", domain=self.root[8:])

        base = self.root + "/Comic/"
        data = {"_extractor": ReadcomiconlineComicExtractor}
        url = f"{self.root}/{path}"
        params = {"page": 1}

        while True:
            page = self.request(url, params=params).text

            for href in text.extract_iter(page, '<a href="/Comic/', '"'):
                if "?id=" not in href:
                    yield Message.Queue, base + href, data

            if ">&rsaquo; Next <" not in page:
                break
            params["page"] += 1


def baeu(url, root="", root_blogspot="https://2.bp.blogspot.com"):
    """https://readcomiconline.li/Scripts/rguard.min.js?v=1.5.4"""
    if not root:
        root = root_blogspot

    url = url.replace("pw_.g28x", "b")
    url = url.replace("d2pr.x_27", "h")

    if url.startswith("https"):
        return url.replace(root_blogspot, root, 1)

    path, sep, query = url.partition("?")

    contains_s0 = "=s0" in path
    path = path[:-3 if contains_s0 else -6]
    path = path[15:33] + path[50:]  # step1()
    path = path[0:-11] + path[-2:]  # step2()
    path = binascii.a2b_base64(path).decode()  # atob()
    path = path[0:13] + path[17:]
    path = path[0:-2] + ("=s0" if contains_s0 else "=s1600")
    return root + "/" + path + sep + query



# -*- coding: utf-8 -*-

# Copyright 2024-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Download Archives"""

import os
import logging
from . import util, formatter

log = logging.getLogger("archive")


def connect(path, prefix, format,
            table=None, mode=None, pragma=None, kwdict=None, cache_key=None):
    keygen = formatter.parse(prefix + format).format_map

    if isinstance(path, str) and path.startswith(
            ("postgres://", "postgresql://")):
        if mode == "memory":
            cls = DownloadArchivePostgresqlMemory
        else:
            cls = DownloadArchivePostgresql
    else:
        path = util.expand_path(path)
        if kwdict is not None and "{" in path:
            path = formatter.parse(path).format_map(kwdict)
        if mode == "memory":
            cls = DownloadArchiveMemory
        else:
            cls = DownloadArchive

    if kwdict is not None and table:
        table = formatter.parse(table).format_map(kwdict)

    return cls(path, keygen, table, pragma, cache_key)


def sanitize(name):
    return f'''"{name.replace('"', '_')}"'''


class DownloadArchive():
    _sqlite3 = None

    def __init__(self, path, keygen, table=None, pragma=None, cache_key=None):
        if self._sqlite3 is None:
            DownloadArchive._sqlite3 = __import__("sqlite3")

        try:
            con = self._sqlite3.connect(
                path, timeout=60, check_same_thread=False)
        except self._sqlite3.OperationalError:
            os.makedirs(os.path.dirname(path))
            con = self._sqlite3.connect(
                path, timeout=60, check_same_thread=False)
        con.isolation_level = None

        self.keygen = keygen
        self.connection = con
        self.close = con.close
        self.cursor = cursor = con.cursor()
        self._cache_key = cache_key or "_archive_key"

        table = "archive" if table is None else sanitize(table)
        self._stmt_select = (
            f"SELECT 1 "
            f"FROM {table} "
            f"WHERE entry=? "
            f"LIMIT 1")
        self._stmt_insert = (
            f"INSERT OR IGNORE INTO {table} "
            f"(entry) VALUES (?)")

        if pragma:
            for stmt in pragma:
                cursor.execute(f"PRAGMA {stmt}")

        try:
            cursor.execute(f"CREATE TABLE IF NOT EXISTS {table} "
                           f"(entry TEXT PRIMARY KEY) WITHOUT ROWID")
        except self._sqlite3.OperationalError:
            # fallback for missing WITHOUT ROWID support (#553)
            cursor.execute(f"CREATE TABLE IF NOT EXISTS {table} "
                           f"(entry TEXT PRIMARY KEY)")

    def add(self, kwdict):
        """Add item described by 'kwdict' to archive"""
        key = kwdict.get(self._cache_key) or self.keygen(kwdict)
        self.cursor.execute(self._stmt_insert, (key,))

    def check(self, kwdict):
        """Return True if the item described by 'kwdict' exists in archive"""
        key = kwdict[self._cache_key] = self.keygen(kwdict)
        self.cursor.execute(self._stmt_select, (key,))
        return self.cursor.fetchone()

    def finalize(self):
        pass


class DownloadArchiveMemory(DownloadArchive):

    def __init__(self, path, keygen, table=None, pragma=None, cache_key=None):
        DownloadArchive.__init__(
            self, path, keygen, table, pragma, cache_key)
        self.keys = set()

    def add(self, kwdict):
        self.keys.add(
            kwdict.get(self._cache_key) or
            self.keygen(kwdict))

    def check(self, kwdict):
        key = kwdict[self._cache_key] = self.keygen(kwdict)
        if key in self.keys:
            return True
        self.cursor.execute(self._stmt_select, (key,))
        return self.cursor.fetchone()

    def finalize(self):
        if not self.keys:
            return

        cursor = self.cursor
        with self.connection:
            try:
                cursor.execute("BEGIN")
            except self._sqlite3.OperationalError:
                pass

            stmt = self._stmt_insert
            if len(self.keys) < 100:
                for key in self.keys:
                    cursor.execute(stmt, (key,))
            else:
                cursor.executemany(stmt, ((key,) for key in self.keys))


class DownloadArchivePostgresql():
    _psycopg = None

    def __init__(self, uri, keygen, table=None, pragma=None, cache_key=None):
        if self._psycopg is None:
            DownloadArchivePostgresql._psycopg = __import__("psycopg")

        self.connection = con = self._psycopg.connect(uri)
        self.cursor = cursor = con.cursor()
        self.close = con.close
        self.keygen = keygen
        self._cache_key = cache_key or "_archive_key"

        table = "archive" if table is None else sanitize(table)
        self._stmt_select = (
            f"SELECT true "
            f"FROM {table} "
            f"WHERE entry=%s "
            f"LIMIT 1")
        self._stmt_insert = (
            f"INSERT INTO {table} (entry) "
            f"VALUES (%s) "
            f"ON CONFLICT DO NOTHING")

        try:
            cursor.execute(f"CREATE TABLE IF NOT EXISTS {table} "
                           f"(entry TEXT PRIMARY KEY)")
            con.commit()
        except Exception as exc:
            log.error("%s: %s when creating '%s' table: %s",
                      con, exc.__class__.__name__, table, exc)
            con.rollback()
            raise

    def add(self, kwdict):
        key = kwdict.get(self._cache_key) or self.keygen(kwdict)
        try:
            self.cursor.execute(self._stmt_insert, (key,))
            self.connection.commit()
        except Exception as exc:
            log.error("%s: %s when writing entry: %s",
                      self.connection, exc.__class__.__name__, exc)
            self.connection.rollback()

    def check(self, kwdict):
        key = kwdict[self._cache_key] = self.keygen(kwdict)
        try:
            self.cursor.execute(self._stmt_select, (key,))
            return self.cursor.fetchone()
        except Exception as exc:
            log.error("%s: %s when checking entry: %s",
                      self.connection, exc.__class__.__name__, exc)
            self.connection.rollback()
            return False

    def finalize(self):
        pass


class DownloadArchivePostgresqlMemory(DownloadArchivePostgresql):

    def __init__(self, path, keygen, table=None, pragma=None, cache_key=None):
        DownloadArchivePostgresql.__init__(
            self, path, keygen, table, pragma, cache_key)
        self.keys = set()

    def add(self, kwdict):
        self.keys.add(
            kwdict.get(self._cache_key) or
            self.keygen(kwdict))

    def check(self, kwdict):
        key = kwdict[self._cache_key] = self.keygen(kwdict)
        if key in self.keys:
            return True
        try:
            self.cursor.execute(self._stmt_select, (key,))
            return self.cursor.fetchone()
        except Exception as exc:
            log.error("%s: %s when checking entry: %s",
                      self.connection, exc.__class__.__name__, exc)
            self.connection.rollback()
            return False

    def finalize(self):
        if not self.keys:
            return
        try:
            self.cursor.executemany(
                self._stmt_insert,
                ((key,) for key in self.keys))
            self.connection.commit()
        except Exception as exc:
            log.error("%s: %s when writing entries: %s",
                      self.connection, exc.__class__.__name__, exc)
            self.connection.rollback()
